import copy
import threading
import typing
import uuid

from nion.swift.model import DataItem
from nion.swift.model import FileStorageSystem
from nion.swift.model import Migration
from nion.swift.model import Utility
from nion.utils import Event


class MemoryStorageHandler:

    def __init__(self, uuid, properties, data, data_read_event):
        self.__uuid = uuid
        self.__properties = properties
        self.__data = data
        self.__data_read_event = data_read_event

    def close(self):
        self.__uuid = None
        self.__properties = None
        self.__data = None

    @property
    def reference(self):
        return str(self.__uuid)

    def read_properties(self):
        return copy.deepcopy(self.__properties.get(self.__uuid, dict()))

    def read_data(self):
        self.__data_read_event.fire(self.__uuid)
        return self.__data.get(self.__uuid)

    def write_properties(self, properties, file_datetime):
        self.__properties[self.__uuid] = Utility.clean_dict(properties)

    def write_data(self, data, file_datetime):
        self.__data[self.__uuid] = data.copy()


class MemoryStorageSystem:
    """Simulate the file system interface for storage.

    data_item_storage is a dict used to hold data item files.

    library_storage is a  dict used to hold the library, while in memory.

    library_persistent_storage is a  dict used to hold the library, after written to disk.
    """

    def __init__(self, *, auto_migrations=None):
        self.data = dict()
        self.__data_item_storage = dict()
        self.trash = dict()
        self._test_data_read_event = Event.Event()
        self.__library_storage = dict()
        self.__library_storage_lock = threading.RLock()
        self.library_persistent_storage_enabled = False
        self.__library_persistent_storage = dict()
        self.__data_item_storage_adapters = dict()
        self.__write_delay_counts = dict()
        self.__auto_migrations = auto_migrations or list()
        for auto_migration in self.__auto_migrations:
            auto_migration.storage_system = self

    def __deepcopy__(self, memo):
        deepcopy = self.__class__()
        deepcopy._set_properties(copy.deepcopy(self.library_storage_properties))
        deepcopy._set_storage_properties(copy.deepcopy(self.persistent_storage_properties))
        deepcopy.data = copy.deepcopy(self.data)
        memo[id(self)] = deepcopy
        return deepcopy

    def reset(self):
        self.__data_item_storage_adapters = dict()

    def get_auto_migrations(self):
        return self.__auto_migrations

    def __write_properties(self, object):
        if self.__write_delay_counts.get(object, 0) == 0:
            persistent_object_parent = object.persistent_object_parent if object else None
            if object and isinstance(object, DataItem.DataItem):
                self.__get_storage_for_item(object).rewrite_item(object)
            elif not persistent_object_parent:
                if self.library_persistent_storage_enabled:
                    self.__library_persistent_storage = copy.deepcopy(self.__library_storage)
            elif persistent_object_parent:
                self.__write_properties(persistent_object_parent.parent)

    @property
    def library_persistent_storage(self):
        return self.__library_persistent_storage

    @property
    def library_storage_properties(self):
        with self.__library_storage_lock:
            return copy.deepcopy(self.__library_storage)

    @property
    def persistent_storage_properties(self):
        return self.__data_item_storage

    def _set_properties(self, properties):
        """Set the properties; used for testing."""
        with self.__library_storage_lock:
            self.__library_storage = properties

    def rewrite_properties(self, properties):
        """Set the properties and write to disk."""
        with self.__library_storage_lock:
            self.__library_storage = properties
        self.__write_properties(None)

    def get_properties(self, object):
        return self.__get_storage_dict(object)

    def __get_storage_dict(self, object):
        persistent_object_parent = object.persistent_object_parent
        if isinstance(object, DataItem.DataItem):
            return self.__get_storage_for_item(object).properties
        if not persistent_object_parent:
            return self.__library_storage
        else:
            parent_storage_dict = self.__get_storage_dict(persistent_object_parent.parent)
            return object.get_accessor_in_parent()(parent_storage_dict)

    def __update_modified_and_get_storage_dict(self, object):
        storage_dict = self.__get_storage_dict(object)
        with self.__library_storage_lock:
            storage_dict["modified"] = object.modified.isoformat()
        persistent_object_parent = object.persistent_object_parent
        parent = persistent_object_parent.parent if persistent_object_parent else None
        if parent:
            self.__update_modified_and_get_storage_dict(parent)
        return storage_dict

    def insert_item(self, parent, name, before_index, item):
        if isinstance(item, DataItem.DataItem):
            item.persistent_object_context = parent.persistent_object_context
            storage_handler = self.make_storage_handler(item)
            self.register_data_item(item, item.uuid, storage_handler, item.write_to_dict())
        else:
            storage_dict = self.__update_modified_and_get_storage_dict(parent)
            with self.__library_storage_lock:
                item_list = storage_dict.setdefault(name, list())
                item_dict = item.write_to_dict()
                item_list.insert(before_index, item_dict)
                item.persistent_object_context = parent.persistent_object_context
            self.__write_properties(parent)

    def remove_item(self, parent, name, index, item):
        if isinstance(item, DataItem.DataItem):
            self.delete_item(item, safe=True)
            item.persistent_object_context = None
            self.unregister_data_item(item)
        else:
            storage_dict = self.__update_modified_and_get_storage_dict(parent)
            with self.__library_storage_lock:
                item_list = storage_dict[name]
                del item_list[index]
            self.__write_properties(parent)
            item.persistent_object_context = None

    def set_item(self, parent, name, item):
        storage_dict = self.__update_modified_and_get_storage_dict(parent)
        with self.__library_storage_lock:
            if item:
                item_dict = item.write_to_dict()
                storage_dict[name] = item_dict
                item.persistent_object_context = parent.persistent_object_context
            else:
                if name in storage_dict:
                    del storage_dict[name]
        self.__write_properties(parent)

    def set_property(self, object, name, value):
        storage_dict = self.__update_modified_and_get_storage_dict(object)
        with self.__library_storage_lock:
            storage_dict[name] = value
        self.__write_properties(object)

    def clear_property(self, object, name):
        storage_dict = self.__update_modified_and_get_storage_dict(object)
        with self.__library_storage_lock:
            storage_dict.pop(name, None)
        self.__write_properties(object)

    def _set_storage_properties(self, properties):
        self.__data_item_storage = properties

    def find_data_items(self):
        storage_handlers = list()
        for key in sorted(self.__data_item_storage):
            self.__data_item_storage[key].setdefault("uuid", str(uuid.uuid4()))
            storage_handlers.append(MemoryStorageHandler(key, self.__data_item_storage, self.data, self._test_data_read_event))
        return storage_handlers

    def make_storage_handler(self, data_item, file_handler=None):
        data_item_uuid_str = str(data_item.uuid)
        return MemoryStorageHandler(data_item_uuid_str, self.__data_item_storage, self.data, self._test_data_read_event)

    def remove_storage_handler(self, storage_handler, *, safe: bool=False) -> None:
        storage_handler_reference = storage_handler.reference
        data = self.data.pop(storage_handler_reference, None)
        properties = self.__data_item_storage.pop(storage_handler_reference)
        if safe:
            assert storage_handler_reference not in self.trash
            self.trash[storage_handler_reference] = {"data": data, "properties": properties}

    def restore_item(self, data_item_uuid: uuid.UUID) -> typing.Tuple[typing.Optional[dict], bool]:
        data_item_uuid_str = str(data_item_uuid)
        trash_entry = self.trash.pop(data_item_uuid_str)
        assert data_item_uuid_str not in self.__data_item_storage
        assert data_item_uuid_str not in self.data
        self.__data_item_storage[data_item_uuid_str] = Migration.transform_to_latest(trash_entry["properties"])
        self.data[data_item_uuid_str] = trash_entry["data"]
        properties = self.__data_item_storage.get(data_item_uuid_str, dict())
        properties["__large_format"] = False
        return properties

    def purge_removed_storage_handlers(self):
        self.trash = dict()

    def prune(self):
        pass

    def register_data_item(self, item, item_uuid, storage_handler, properties: dict) -> None:
        assert item_uuid not in self.__data_item_storage_adapters
        storage = FileStorageSystem.DataItemStorageAdapter(self, storage_handler, properties)
        self.__data_item_storage_adapters[item_uuid] = storage
        if item and self.is_write_delayed(item):
            storage.set_write_delayed(item, True)

    def unregister_data_item(self, item: DataItem) -> None:
        assert item.uuid in self.__data_item_storage_adapters
        self.__data_item_storage_adapters.pop(item.uuid).close()

    def __get_storage_for_item(self, item: DataItem) -> FileStorageSystem.DataItemStorageAdapter:
        if not item.uuid in self.__data_item_storage_adapters:
            storage_handler = self.make_storage_handler(item)
            self.register_data_item(item, item.uuid, storage_handler, item.write_to_dict())
        return self.__data_item_storage_adapters.get(item.uuid)

    def get_storage_property(self, data_item: DataItem.DataItem, name: str) -> typing.Optional[str]:
        if name == "file_path":
            storage = self.__get_storage_for_item(data_item)
            return storage._storage_handler.reference if storage else None
        return None

    def read_external_data(self, item, name):
        if isinstance(item, DataItem.DataItem) and name == "data":
            storage = self.__get_storage_for_item(item)
            return storage.load_data(item)
        return None

    def write_external_data(self, item, name, value) -> None:
        if isinstance(item, DataItem.DataItem) and name == "data":
            storage = self.__get_storage_for_item(item)
            storage.update_data(item, value)

    def delete_item(self, data_item, safe: bool=False) -> None:
        storage = self.__get_storage_for_item(data_item)
        self.remove_storage_handler(storage._storage_handler, safe=safe)

    def enter_write_delay(self, object) -> None:
        count = self.__write_delay_counts.setdefault(object, 0)
        if count == 0:
            if isinstance(object, DataItem.DataItem):
                self.set_write_delayed(object, True)
        self.__write_delay_counts[object] = count + 1

    def exit_write_delay(self, object) -> None:
        count = self.__write_delay_counts.get(object, 1)
        count -= 1
        if count == 0:
            if isinstance(object, DataItem.DataItem):
                self.set_write_delayed(object, False)
            self.__write_delay_counts.pop(object)
        else:
            self.__write_delay_counts[object] = count

    def set_write_delayed(self, data_item, write_delayed: bool) -> None:
        storage = self.__data_item_storage_adapters.get(data_item.uuid)
        if storage:
            storage.set_write_delayed(data_item, write_delayed)

    def is_write_delayed(self, data_item) -> bool:
        if isinstance(data_item, DataItem.DataItem):
            return self.__write_delay_counts.get(data_item, 0) > 0
        return False

    def rewrite_item(self, item) -> None:
        if isinstance(item, DataItem.DataItem):
            storage = self.__get_storage_for_item(item)
            storage.rewrite_item(item)
        else:
            self.__write_properties(item)

    def read_data_items_version_stats(self):
        return FileStorageSystem.read_data_items_version_stats(self)

    def read_library(self, ignore_older_files) -> typing.Mapping:
        return FileStorageSystem.read_library(self, ignore_older_files)



# standard libraries
import asyncio
import copy
import functools
import gettext
import logging
import pkgutil
import threading
import uuid

# third party libraries
# None

# local libraries
from nion.swift import MimeTypes
from nion.swift import Panel
from nion.swift import Thumbnails
from nion.swift.model import DataGroup
from nion.swift.model import DataItem
from nion.swift.model import DisplayItem
from nion.ui import CanvasItem
from nion.ui import DrawingContext
from nion.ui import GridCanvasItem
from nion.ui import ListCanvasItem
from nion.ui import Widgets
from nion.utils import Event
from nion.utils import Geometry
from nion.utils import ListModel

_ = gettext.gettext


"""
    The data panel has two parts:

    (1) a selection of what collection is displayed, which may be a data group, smart data group,
    or the whole document. If the whole document, then an optional filter may also be applied.
    "within the last 24 hours" would be an example filter.

    (2) the list of display items from the collection. the user may further refine the list of
    items by filtering by additional criteria. the user also chooses the sorting on the list of
    display items.

"""


# support the 'count' display for data groups. count the display items that are children of the container (which
# can be a data group or a document controller) and also display items in all of their child groups.
def get_display_item_count_flat(document_model, container) -> int:

    def append_display_item_flat(container, display_items):
        if isinstance(container, DataItem.DataItem):
            display_items.append(document_model.get_display_item_for_data_item(container))
        if hasattr(container, "data_items"):
            for data_item in container.data_items:
                append_display_item_flat(data_item, display_items)

    display_items = []
    append_display_item_flat(container, display_items)
    return len(display_items)


class DisplayItemAdapter:
    """ Provide a simplified interface to a display item for the purpose of display.

        The display_item property is always valid.

        There is typically one display item associated with each display item.

        Provides the following interface:
            (method) close()
            (property, read-only) display_item
            (property, read-only) title_str
            (property, read-only) datetime_str
            (property, read-only) format_str
            (property, read-only) status_str
            (method) drag_started(ui, x, y, modifiers), returns mime_data, thumbnail_data
            (event) needs_update_event
    """

    def __init__(self, display_item: DisplayItem.DisplayItem, ui):
        self.__display_item = display_item
        self.ui = ui
        self.needs_update_event = Event.Event()

        def display_item_changed():
            self.needs_update_event.fire()

        self.__display_changed_event_listener = display_item.item_changed_event.listen(display_item_changed) if display_item else None

        self.__thumbnail_updated_event_listener = None
        self.__thumbnail_source = None

    def close(self):
        # remove the listener.
        if self.__thumbnail_updated_event_listener:
            self.__thumbnail_updated_event_listener.close()
            self.__thumbnail_updated_event_listener = None
        if self.__thumbnail_source:
            self.__thumbnail_source.close()
            self.__thumbnail_source = None
        if self.__display_changed_event_listener:
            self.__display_changed_event_listener.close()
            self.__display_changed_event_listener = None

    @property
    def display_item(self) -> DisplayItem.DisplayItem:
        return self.__display_item

    @property
    def data_item(self) -> DataItem.DataItem:
        return self.__display_item.data_item if self.__display_item else None

    def __create_thumbnail_source(self):
        # grab the display specifier and if there is a display, handle thumbnail updating.
        if self.__display_item and not self.__thumbnail_source:
            self.__thumbnail_source = Thumbnails.ThumbnailManager().thumbnail_source_for_display_item(self.ui, self.__display_item)

            def thumbnail_updated():
                self.needs_update_event.fire()

            self.__thumbnail_updated_event_listener = self.__thumbnail_source.thumbnail_updated_event.listen(thumbnail_updated)

    def __create_thumbnail(self, draw_rect):
        drawing_context = DrawingContext.DrawingContext()
        if self.__display_item:
            self.__create_thumbnail_source()
            thumbnail_data = self.__thumbnail_source.thumbnail_data
            if thumbnail_data is not None:
                draw_rect = Geometry.fit_to_size(draw_rect, thumbnail_data.shape)
                drawing_context.draw_image(thumbnail_data, draw_rect[0][1], draw_rect[0][0], draw_rect[1][1], draw_rect[1][0])
        return drawing_context

    @property
    def title_str(self) -> str:
        return self.__display_item.displayed_title if self.__display_item else str()

    @property
    def format_str(self) -> str:
        return self.__display_item.size_and_data_format_as_string if self.__display_item else str()

    @property
    def datetime_str(self) -> str:
        return self.__display_item.date_for_sorting_local_as_string if self.__display_item else str()

    @property
    def status_str(self) -> str:
        return self.__display_item.status_str if self.__display_item else str()

    def drag_started(self, ui, x, y, modifiers):
        if self.__display_item:
            mime_data = self.ui.create_mime_data()
            if self.__display_item:
                mime_data.set_data_as_string(MimeTypes.DISPLAY_ITEM_MIME_TYPE, str(self.__display_item.uuid))
            self.__create_thumbnail_source()
            thumbnail_data = self.__thumbnail_source.thumbnail_data if self.__thumbnail_source else None
            return mime_data, thumbnail_data
        return None, None

    def draw_list_item(self, drawing_context, rect):
        with drawing_context.saver():
            draw_rect = ((rect[0][0] + 4, rect[0][1] + 4), (72, 72))
            drawing_context.add(self.__create_thumbnail(draw_rect))
            drawing_context.fill_style = "#000"
            drawing_context.font = "11px serif"
            drawing_context.fill_text(self.title_str, rect[0][1] + 4 + 72 + 4, rect[0][0] + 4 + 12)
            drawing_context.fill_text(self.format_str, rect[0][1] + 4 + 72 + 4, rect[0][0] + 4 + 12 + 15)
            drawing_context.fill_text(self.datetime_str, rect[0][1] + 4 + 72 + 4, rect[0][0] + 4 + 12 + 15 + 15)
            drawing_context.fill_text(self.status_str, rect[0][1] + 4 + 72 + 4, rect[0][0] + 4 + 12 + 15 + 15 + 15)

    def draw_grid_item(self, drawing_context, rect):
        drawing_context.add(self.__create_thumbnail(rect.inset(6)))


class DataListController:
    """Control a list of display items in a list widget.

    The following properties are available:
        selected_indexes (r/o)

    The following methods can be called:
        close()

    The controller provides the following callbacks:
        on_delete_display_item_adapters(display_item_adapters)
        on_key_pressed(key)
        on_display_item_adapter_selection_changed(display_item_adapters)
        on_focus_changed(focused)
        on_context_menu_event(display_item_adapter, x, y, gx, gy)

    Display items should respond to these properties and methods and events:
        (method) close()
        (property, read-only) title_str
        (property, read-only) datetime_str
        (property, read-only) format_str
        (property, read-only) status_str
        (method) draw_list_item(drawing_context, draw_rect)
        (method) drag_started(ui, x, y, modifiers), returns mime_data, thumbnail_data
    """

    def __init__(self, event_loop: asyncio.AbstractEventLoop, ui, display_item_adapters_model, selection):
        super().__init__()
        self.__event_loop = event_loop
        self.__pending_tasks = list()
        self.ui = ui
        self.__selection = selection
        self.on_delete_display_item_adapters = None
        self.on_key_pressed = None

        self.__display_item_adapters = list()
        self.__display_item_adapter_needs_update_listeners = list()

        self.__display_item_adapters_model = display_item_adapters_model
        self.__display_item_adapter_inserted_event_listener = self.__display_item_adapters_model.item_inserted_event.listen(self.__display_item_adapter_inserted)
        self.__display_item_adapter_removed_event_listener = self.__display_item_adapters_model.item_removed_event.listen(self.__display_item_adapter_removed)

        class ListCanvasItemDelegate:
            def __init__(self, data_list_controller):
                self.__data_list_controller = data_list_controller

            @property
            def item_count(self):
                return self.__data_list_controller.display_item_adapter_count

            @property
            def items(self):
                return self.__data_list_controller.display_item_adapters

            def paint_item(self, drawing_context, display_item_adapter, rect, is_selected):
                display_item_adapter.draw_list_item(drawing_context, rect)

            def on_context_menu_event(self, index, x, y, gx, gy):
                return self.__data_list_controller.context_menu_event(index, x, y, gx, gy)

            def on_delete_pressed(self):
                self.__data_list_controller._delete_pressed()

            def on_key_pressed(self, key):
                return self.__data_list_controller._key_pressed(key)

            def on_drag_started(self, index, x, y, modifiers):
                self.__data_list_controller.drag_started(index, x, y, modifiers)

        self.__list_canvas_item = ListCanvasItem.ListCanvasItem(ListCanvasItemDelegate(self), self.__selection)
        def focus_changed(focused):
            self.__list_canvas_item.update()
            if self.on_focus_changed:
                self.on_focus_changed(focused)
        self.__list_canvas_item.on_focus_changed = focus_changed
        self.scroll_area_canvas_item = CanvasItem.ScrollAreaCanvasItem(self.__list_canvas_item)
        self.scroll_area_canvas_item.auto_resize_contents = True
        self.scroll_bar_canvas_item = CanvasItem.ScrollBarCanvasItem(self.scroll_area_canvas_item)
        self.scroll_group_canvas_item = CanvasItem.CanvasItemComposition()
        self.scroll_group_canvas_item.layout = CanvasItem.CanvasItemRowLayout()
        self.scroll_group_canvas_item.add_canvas_item(self.scroll_area_canvas_item)
        self.scroll_group_canvas_item.add_canvas_item(self.scroll_bar_canvas_item)
        self.canvas_item = self.scroll_group_canvas_item
        def selection_changed():
            self.selected_indexes = list(self.__selection.indexes)
            if callable(self.on_display_item_adapter_selection_changed):
                self.on_display_item_adapter_selection_changed([self.__display_item_adapters[index] for index in list(self.__selection.indexes)])
            self.__list_canvas_item.make_selection_visible()
        self.__selection_changed_listener = self.__selection.changed_event.listen(selection_changed)
        self.selected_indexes = list()
        self.on_display_item_adapter_selection_changed = None
        self.on_context_menu_event = None
        self.on_focus_changed = None
        self.on_drag_started = None

        # changed display items keep track of items whose content has changed
        # the content changed messages may come from a thread so have to be
        # moved to the main thread via this object.
        self.__changed_display_item_adapters = False
        self.__changed_display_item_adapters_mutex = threading.RLock()

        for index, display_item_adapter in enumerate(self.__display_item_adapters_model.display_item_adapters):
            self.__display_item_adapter_inserted("display_item_adapters", display_item_adapter, index)

    def close(self):
        for pending_task in self.__pending_tasks:
            pending_task.cancel()
        self.__pending_tasks = None
        self.__selection_changed_listener.close()
        self.__selection_changed_listener = None
        for display_item_adapter_needs_update_listener in self.__display_item_adapter_needs_update_listeners:
            display_item_adapter_needs_update_listener.close()
        self.__display_item_adapter_needs_update_listeners = None
        self.__display_item_adapter_inserted_event_listener.close()
        self.__display_item_adapter_inserted_event_listener = None
        self.__display_item_adapter_removed_event_listener.close()
        self.__display_item_adapter_removed_event_listener = None
        self.__display_item_adapters = None
        self.on_display_item_adapter_selection_changed = None
        self.on_context_menu_event = None
        self.on_drag_started = None
        self.on_focus_changed = None
        self.on_delete_display_item_adapters = None
        self.on_key_pressed = None

    async def __update_display_item_adapters(self):
        # handle the 'changed' stuff. a call to this function is scheduled
        # whenever __changed_display_item_adapters changes.
        with self.__changed_display_item_adapters_mutex:
            changed_display_item_adapters = self.__changed_display_item_adapters
            self.__changed_display_item_adapters = False
        if changed_display_item_adapters:
            self.__list_canvas_item.update()
        self.__pending_tasks.pop(0)

    def make_selection_visible(self):
        self.__list_canvas_item.make_selection_visible()

    # this message comes from the canvas item when delete key is pressed
    def _delete_pressed(self):
        if callable(self.on_delete_display_item_adapters):
            self.on_delete_display_item_adapters([self.__display_item_adapters[index] for index in self.__selection.indexes])

    # this message comes from the canvas item when a key is pressed
    def _key_pressed(self, key):
        if callable(self.on_key_pressed):
            return self.on_key_pressed(key)
        return False

    @property
    def display_item_adapter_count(self):
        return len(self.__display_item_adapters)

    @property
    def display_item_adapters(self):
        return copy.copy(self.__display_item_adapters)

    def _test_get_display_item_adapter(self, index):
        return self.__display_item_adapters[index]

    def context_menu_event(self, index, x, y, gx, gy):
        if self.on_context_menu_event:
            display_item_adapter = self.__display_item_adapters[index] if index is not None else None
            return self.on_context_menu_event(display_item_adapter, x, y, gx, gy)
        return False

    def drag_started(self, index, x, y, modifiers):
        mime_data, thumbnail_data = self.__display_item_adapters[index].drag_started(self.ui, x, y, modifiers)
        if mime_data:
            if self.on_drag_started:
                self.on_drag_started(mime_data, thumbnail_data)

    def __display_item_adapter_needs_update(self):
        with self.__changed_display_item_adapters_mutex:
            self.__changed_display_item_adapters = True
            self.__pending_tasks.append(self.__event_loop.create_task(self.__update_display_item_adapters()))

    # call this method to insert a display item
    # not thread safe
    def __display_item_adapter_inserted(self, key, display_item_adapter, before_index):
        if key == "display_item_adapters":
            self.__display_item_adapters.insert(before_index, display_item_adapter)
            self.__display_item_adapter_needs_update_listeners.insert(before_index, display_item_adapter.needs_update_event.listen(self.__display_item_adapter_needs_update))
            # tell the icon view to update.
            self.__list_canvas_item.refresh_layout()
            self.__list_canvas_item.update()

    # call this method to remove a display item (by index)
    # not thread safe
    def __display_item_adapter_removed(self, key, display_item_adapter, index):
        if key == "display_item_adapters":
            self.__display_item_adapter_needs_update_listeners[index].close()
            del self.__display_item_adapter_needs_update_listeners[index]
            del self.__display_item_adapters[index]
            self.__list_canvas_item.refresh_layout()
            self.__list_canvas_item.update()


class DataGridController:
    """Control a grid of display items in a grid widget.

    The following properties are available:
        selected_indexes (r/o)

    The following methods can be called:
        close()

    The controller provides the following callbacks:
        on_delete_display_item_adapters(display_item_adapters)
        on_key_pressed(key)
        on_display_item_adapter_selection_changed(display_item_adapters)
        on_display_item_adapter_double_clicked(display_item_adapter)
        on_focus_changed(focused)
        on_context_menu_event(display_item_adapter, x, y, gx, gy)
        on_drag_started(mime_data, thumbnail_data)

    Display items should respond to these properties and methods and events:
        (method) close()
        (property, read-only) thumbnail
        (property, read-only) title_str
        (property, read-only) datetime_str
        (property, read-only) format_str
        (property, read-only) status_str
        (method) draw_grid_item(drawing_context, draw_rect)
        (method) drag_started(ui, x, y, modifiers), returns mime_data, thumbnail_data
    """

    def __init__(self, event_loop: asyncio.AbstractEventLoop, ui, display_item_adapters_model, selection, direction=GridCanvasItem.Direction.Row, wrap=True):
        super().__init__()
        self.__event_loop = event_loop
        self.__pending_tasks = list()
        self.ui = ui
        self.__selection = selection
        self.on_delete_display_item_adapters = None
        self.on_key_pressed = None
        self.on_display_item_adapter_double_clicked = None
        self.on_display_item_adapter_selection_changed = None
        self.on_context_menu_event = None
        self.on_focus_changed = None
        self.on_drag_started = None

        self.__display_item_adapters = list()
        self.__display_item_adapter_needs_update_listeners = list()

        self.__display_item_adapters_model = display_item_adapters_model
        self.__display_item_adapter_inserted_event_listener = self.__display_item_adapters_model.item_inserted_event.listen(self.__display_item_adapter_inserted)
        self.__display_item_adapter_removed_event_listener = self.__display_item_adapters_model.item_removed_event.listen(self.__display_item_adapter_removed)

        class GridCanvasItemDelegate:
            def __init__(self, data_grid_controller):
                self.__data_grid_controller = data_grid_controller

            @property
            def item_count(self):
                return self.__data_grid_controller.display_item_adapter_count

            @property
            def items(self):
                return self.__data_grid_controller.display_item_adapters

            def paint_item(self, drawing_context, display_item_adapter, rect, is_selected):
                display_item_adapter.draw_grid_item(drawing_context, rect)

            def on_context_menu_event(self, index, x, y, gx, gy):
                return self.__data_grid_controller.context_menu_event(index, x, y, gx, gy)

            def on_delete_pressed(self):
                self.__data_grid_controller._delete_pressed()

            def on_key_pressed(self, key):
                return self.__data_grid_controller._key_pressed(key)

            def on_mouse_double_clicked(self, mouse_index, x, y, modifiers):
                return self.__data_grid_controller._double_clicked()

            def on_drag_started(self, index, x, y, modifiers):
                self.__data_grid_controller.drag_started(index, x, y, modifiers)

        self.icon_view_canvas_item = GridCanvasItem.GridCanvasItem(GridCanvasItemDelegate(self), self.__selection, direction, wrap)
        def icon_view_canvas_item_focus_changed(focused):
            self.icon_view_canvas_item.update()
            if self.on_focus_changed:
                self.on_focus_changed(focused)
        self.icon_view_canvas_item.on_focus_changed = icon_view_canvas_item_focus_changed
        self.scroll_area_canvas_item = CanvasItem.ScrollAreaCanvasItem(self.icon_view_canvas_item)
        self.scroll_area_canvas_item.auto_resize_contents = True
        self.scroll_group_canvas_item = CanvasItem.CanvasItemComposition()
        if (wrap and direction == GridCanvasItem.Direction.Row) or (not wrap and direction == GridCanvasItem.Direction.Column):
            self.scroll_bar_canvas_item = CanvasItem.ScrollBarCanvasItem(self.scroll_area_canvas_item)
            self.scroll_group_canvas_item.layout = CanvasItem.CanvasItemRowLayout()
        else:
            self.scroll_bar_canvas_item = CanvasItem.ScrollBarCanvasItem(self.scroll_area_canvas_item, CanvasItem.Orientation.Horizontal)
            self.scroll_group_canvas_item.layout = CanvasItem.CanvasItemColumnLayout()
        self.scroll_group_canvas_item.add_canvas_item(self.scroll_area_canvas_item)
        self.scroll_group_canvas_item.add_canvas_item(self.scroll_bar_canvas_item)
        if False:  # dual scroll bars, leave here for easy testing
            self.vertical_scroll_bar_canvas_item = CanvasItem.ScrollBarCanvasItem(self.scroll_area_canvas_item)
            self.horizontal_scroll_bar_canvas_item = CanvasItem.ScrollBarCanvasItem(self.scroll_area_canvas_item, CanvasItem.Orientation.Horizontal)
            self.scroll_group_canvas_item.layout = CanvasItem.CanvasItemGridLayout(Geometry.IntSize(width=2, height=2))
            self.scroll_group_canvas_item.add_canvas_item(self.scroll_area_canvas_item, Geometry.IntPoint(x=0, y=0))
            self.scroll_group_canvas_item.add_canvas_item(self.vertical_scroll_bar_canvas_item, Geometry.IntPoint(x=1, y=0))
            self.scroll_group_canvas_item.add_canvas_item(self.horizontal_scroll_bar_canvas_item, Geometry.IntPoint(x=0, y=1))

        self.canvas_item = self.scroll_group_canvas_item

        def selection_changed():
            self.selected_indexes = list(self.__selection.indexes)
            if callable(self.on_display_item_adapter_selection_changed):
                self.on_display_item_adapter_selection_changed([self.__display_item_adapters[index] for index in list(self.__selection.indexes)])
            self.icon_view_canvas_item.make_selection_visible()

        self.__selection_changed_listener = self.__selection.changed_event.listen(selection_changed)
        self.selected_indexes = list()

        # changed display items keep track of items whose content has changed
        # the content changed messages may come from a thread so have to be
        # moved to the main thread via this object.
        self.__changed_display_item_adapters = False
        self.__changed_display_item_adapters_mutex = threading.RLock()
        self.__closed = False

        for index, display_item_adapter in enumerate(self.__display_item_adapters_model.display_item_adapters):
            self.__display_item_adapter_inserted("display_item_adapters", display_item_adapter, index)

    def close(self):
        assert not self.__closed
        for pending_task in self.__pending_tasks:
            pending_task.cancel()
        self.__pending_tasks = None
        self.icon_view_canvas_item.detach_delegate()
        self.__selection_changed_listener.close()
        self.__selection_changed_listener = None
        self.__display_item_adapter_inserted_event_listener.close()
        self.__display_item_adapter_inserted_event_listener = None
        self.__display_item_adapter_removed_event_listener.close()
        self.__display_item_adapter_removed_event_listener = None
        for display_item_adapter_needs_update_listener in self.__display_item_adapter_needs_update_listeners:
            display_item_adapter_needs_update_listener.close()
        self.__display_item_adapter_needs_update_listeners = None
        self.__display_item_adapters = None
        self.on_display_item_adapter_selection_changed = None
        self.on_context_menu_event = None
        self.on_drag_started = None
        self.on_focus_changed = None
        self.on_delete_display_item_adapters = None
        self.on_key_pressed = None
        self.on_display_item_adapter_double_clicked = None
        self.__closed = True

    async def __update_display_item_adapters(self):
        with self.__changed_display_item_adapters_mutex:
            changed_display_item_adapters = self.__changed_display_item_adapters
            self.__changed_display_item_adapters = False
        if changed_display_item_adapters:
            self.icon_view_canvas_item.update()
        self.__pending_tasks.pop(0)

    def clear_selection(self):
        self.__selection.clear()

    def make_selection_visible(self):
        self.icon_view_canvas_item.make_selection_visible()

    # this message comes from the canvas item when delete key is pressed
    def _delete_pressed(self):
        if callable(self.on_delete_display_item_adapters):
            self.on_delete_display_item_adapters([self.__display_item_adapters[index] for index in self.__selection.indexes])

    # this message comes from the canvas item when a key is pressed
    def _key_pressed(self, key):
        if callable(self.on_key_pressed):
            return self.on_key_pressed(key)
        return False

    # this message comes from the canvas item when a key is pressed
    def _double_clicked(self):
        if len(self.__selection.indexes) == 1:
            if callable(self.on_display_item_adapter_double_clicked):
                return self.on_display_item_adapter_double_clicked(self.__display_item_adapters[list(self.__selection.indexes)[0]])
        return False

    @property
    def display_item_adapter_count(self):
        return len(self.__display_item_adapters)

    @property
    def display_item_adapters(self):
        return copy.copy(self.__display_item_adapters)

    def context_menu_event(self, index, x, y, gx, gy):
        if self.on_context_menu_event:
            display_item_adapter = self.__display_item_adapters[index] if index is not None else None
            display_item = display_item_adapter.display_item if display_item_adapter else None
            return self.on_context_menu_event(display_item, x, y, gx, gy)
        return False

    def drag_started(self, index, x, y, modifiers):
        mime_data, thumbnail_data = self.__display_item_adapters[index].drag_started(self.ui, x, y, modifiers)
        if mime_data:
            if self.on_drag_started:
                self.on_drag_started(mime_data, thumbnail_data)

    def __display_item_adapter_needs_update(self):
        with self.__changed_display_item_adapters_mutex:
            self.__changed_display_item_adapters = True
            self.__pending_tasks.append(self.__event_loop.create_task(self.__update_display_item_adapters()))

    # call this method to insert a display item
    # not thread safe
    def __display_item_adapter_inserted(self, key, display_item_adapter, before_index):
        if key == "display_item_adapters":
            self.__display_item_adapters.insert(before_index, display_item_adapter)
            self.__display_item_adapter_needs_update_listeners.insert(before_index, display_item_adapter.needs_update_event.listen(self.__display_item_adapter_needs_update))
            # tell the icon view to update.
            self.icon_view_canvas_item.refresh_layout()
            self.icon_view_canvas_item.update()

    # call this method to remove a display item (by index)
    # not thread safe
    def __display_item_adapter_removed(self, key, display_item_adapter, index):
        if key == "display_item_adapters":
            self.__display_item_adapter_needs_update_listeners[index].close()
            del self.__display_item_adapter_needs_update_listeners[index]
            del self.__display_item_adapters[index]
            self.icon_view_canvas_item.refresh_layout()
            self.icon_view_canvas_item.update()


class DataListWidget(Widgets.CompositeWidgetBase):

    def __init__(self, ui, data_list_controller):
        super().__init__(ui.create_column_widget())
        self.data_list_controller = data_list_controller
        data_list_widget = ui.create_canvas_widget()
        data_list_widget.canvas_item.add_canvas_item(data_list_controller.canvas_item)
        self.content_widget.add(data_list_widget)

        def data_list_drag_started(mime_data, thumbnail_data):
            self.drag(mime_data, thumbnail_data)

        data_list_controller.on_drag_started = data_list_drag_started

    def close(self):
        self.data_list_controller.on_drag_started = None
        self.data_list_controller = None
        super().close()


class DataGridWidget(Widgets.CompositeWidgetBase):

    def __init__(self, ui, data_grid_controller):
        super().__init__(ui.create_column_widget())
        self.data_grid_controller = data_grid_controller
        data_grid_widget = ui.create_canvas_widget()
        data_grid_widget.canvas_item.add_canvas_item(data_grid_controller.canvas_item)
        self.content_widget.add(data_grid_widget)

        def data_list_drag_started(mime_data, thumbnail_data):
            self.drag(mime_data, thumbnail_data)

        data_grid_controller.on_drag_started = data_list_drag_started

    def close(self):
        self.data_grid_controller.on_drag_started = None
        self.data_grid_controller = None
        super().close()


class LibraryModelController:
    """Controller for a list of top level library items."""

    def __init__(self, ui, item_controllers):
        """
        item_controllers is a list of objects that have a title property and a on_title_changed callback that gets
        invoked (on the ui thread) when the title changes externally.
        """
        self.ui = ui
        self.item_model_controller = self.ui.create_item_model_controller(["display"])
        self.item_model_controller.on_can_drop_mime_data = self.can_drop_mime_data
        self.item_model_controller.on_item_drop_mime_data = self.item_drop_mime_data
        self.item_model_controller.supported_drop_actions = self.item_model_controller.DRAG | self.item_model_controller.DROP
        self.item_model_controller.mime_types_for_drop = ["text/uri-list"]
        self.on_receive_files = None
        self.__item_controllers = list()
        self.__item_count = 0
        # build the items
        for item_controller in item_controllers:
            self.__append_item_controller(item_controller)

    def close(self):
        self.__item_controllers = None
        self.item_model_controller.close()
        self.item_model_controller = None
        self.on_receive_files = None

    # not thread safe. must be called on ui thread.
    def __append_item_controller(self, item_controller):
        parent_item = self.item_model_controller.root
        self.item_model_controller.begin_insert(self.__item_count, self.__item_count, parent_item.row, parent_item.id)
        item = self.item_model_controller.create_item()
        parent_item.insert_child(self.__item_count, item)
        self.item_model_controller.end_insert()
        # not thread safe. must be called on ui thread.
        def title_changed(title):
            item.data["display"] = title
            self.item_model_controller.data_changed(item.row, item.parent.row, item.parent.id)
        item_controller.on_title_changed = title_changed
        title_changed(item_controller.title)
        self.__item_controllers.append(item_controller)
        self.__item_count += 1

    def can_drop_mime_data(self, mime_data, action, row, parent_row, parent_id):
        if mime_data.has_file_paths:
            return row < 0  # only accept drops ONTO items, not BETWEEN items
        return False

    def item_drop_mime_data(self, mime_data, action, row, parent_row, parent_id):
        if mime_data.has_file_paths:
            if row >= 0:  # only accept drops ONTO items, not BETWEEN items
                return self.item_model_controller.NONE
            if callable(self.on_receive_files):
                if self.on_receive_files(mime_data.file_paths):
                    return self.item_model_controller.COPY
        return self.item_model_controller.NONE


class DataGroupModelController:
    """ A tree model of the data groups. this class watches for changes to the data groups contained in the document
    controller and responds by updating the item model controller associated with the data group tree view widget. it
    also handles drag and drop and keeps the current selection synchronized with the image panel. """

    def __init__(self, ui, document_controller):
        self.ui = ui
        self.__document_controller = document_controller
        self.__document_model = document_controller.document_model
        self.item_model_controller = self.ui.create_item_model_controller(["display", "edit"])
        self.item_model_controller.on_item_set_data = self.item_set_data
        self.item_model_controller.on_can_drop_mime_data = self.can_drop_mime_data
        self.item_model_controller.on_item_drop_mime_data = self.item_drop_mime_data
        self.item_model_controller.on_item_mime_data = self.item_mime_data
        self.item_model_controller.on_remove_rows = self.remove_rows
        self.item_model_controller.supported_drop_actions = self.item_model_controller.DRAG | self.item_model_controller.DROP
        self.item_model_controller.mime_types_for_drop = ["text/uri-list", MimeTypes.DISPLAY_ITEM_MIME_TYPE, MimeTypes.DATA_GROUP_MIME_TYPE]
        self.__document_model_item_inserted_listener = self.__document_model.item_inserted_event.listen(functools.partial(self.__item_inserted, self.__document_model))
        self.__document_model_item_removed_listener = self.__document_model.item_removed_event.listen(functools.partial(self.__item_removed, self.__document_model))
        self.__mapping = { self.__document_model: self.item_model_controller.root }
        self.on_receive_files = None
        self.on_item_count_changed = None
        # add items that already exist
        self.__data_group_property_changed_listeners = dict()
        self.__data_group_item_inserted_listeners = dict()
        self.__data_group_item_removed_listeners = dict()
        self.__data_group_display_item_inserted_listeners = dict()
        self.__data_group_display_item_removed_listeners = dict()
        data_groups = self.__document_model.data_groups
        for index, data_group in enumerate(data_groups):
            self.__item_inserted(self.__document_model, "data_groups", data_group, index)

    def close(self):
        # cheap way to unlisten to everything
        for object in self.__mapping.keys():
            if isinstance(object, DataGroup.DataGroup):
                self.__data_group_display_item_inserted_listeners[object.uuid].close()
                del self.__data_group_display_item_inserted_listeners[object.uuid]
                self.__data_group_display_item_removed_listeners[object.uuid].close()
                del self.__data_group_display_item_removed_listeners[object.uuid]
                self.__data_group_item_inserted_listeners[object.uuid].close()
                del self.__data_group_item_inserted_listeners[object.uuid]
                self.__data_group_item_removed_listeners[object.uuid].close()
                del self.__data_group_item_removed_listeners[object.uuid]
                self.__data_group_property_changed_listeners[object.uuid].close()
                del self.__data_group_property_changed_listeners[object.uuid]
        self.__document_model_item_inserted_listener.close()
        self.__document_model_item_inserted_listener = None
        self.__document_model_item_removed_listener.close()
        self.__document_model_item_removed_listener = None
        self.item_model_controller.close()
        self.item_model_controller = None
        self.on_receive_files = None
        self.on_item_count_changed = None

    def log(self, parent_id=-1, indent=""):
        parent_id = parent_id if parent_id >= 0 else self.item_model_controller.root.id
        for index, child in enumerate(self.item_model_controller.item_from_id(parent_id).children):
            value = child.data["display"] if "display" in child.data else "---"
            logging.debug(indent + str(index) + ": (" + str(child.id) + ") " + value)
            self.log(child.id, indent + "  ")

    def __get_display_item_count_flat(self, container) -> int:
        return get_display_item_count_flat(self.__document_model, container)

    # this message is received when a display item is inserted into one of the
    # groups we're observing.
    def __item_inserted(self, container, key, object, before_index):
        if key == "data_groups":
            # manage the item model
            parent_item = self.__mapping[container]
            self.item_model_controller.begin_insert(before_index, before_index, parent_item.row, parent_item.id)
            count = self.__get_display_item_count_flat(object)
            properties = {
                "display": str(object) + (" (%i)" % count),
                "edit": object.title,
                "data_group": object
            }
            item = self.item_model_controller.create_item(properties)
            parent_item.insert_child(before_index, item)
            self.__mapping[object] = item
            def property_changed(key):
                if key == "title":
                    self.__update_item_count(object)
            self.__data_group_property_changed_listeners[object.uuid] = object.property_changed_event.listen(property_changed)
            self.__data_group_item_inserted_listeners[object.uuid] = object.item_inserted_event.listen(functools.partial(self.__item_inserted, object))
            self.__data_group_item_removed_listeners[object.uuid] = object.item_removed_event.listen(functools.partial(self.__item_removed, object))
            self.__data_group_display_item_inserted_listeners[object.uuid] = object.display_item_inserted_event.listen(self.__display_item_inserted)
            self.__data_group_display_item_removed_listeners[object.uuid] = object.display_item_removed_event.listen(self.__display_item_removed)
            self.item_model_controller.end_insert()
            # recursively insert items that already exist
            data_groups = object.data_groups
            for index, child_data_group in enumerate(data_groups):
                self.__item_inserted(object, "data_groups", child_data_group, index)
            if callable(self.on_item_count_changed):
                self.on_item_count_changed()

    # this message is received when a display item is removed from one of the
    # groups we're observing.
    def __item_removed(self, container, key, object, index):
        if key == "data_groups":
            assert isinstance(object, DataGroup.DataGroup)
            # get parent and item
            parent_item = self.__mapping[container]
            # manage the item model
            self.item_model_controller.begin_remove(index, index, parent_item.row, parent_item.id)
            self.__data_group_display_item_inserted_listeners[object.uuid].close()
            del self.__data_group_display_item_inserted_listeners[object.uuid]
            self.__data_group_display_item_removed_listeners[object.uuid].close()
            del self.__data_group_display_item_removed_listeners[object.uuid]
            self.__data_group_item_inserted_listeners[object.uuid].close()
            del self.__data_group_item_inserted_listeners[object.uuid]
            self.__data_group_item_removed_listeners[object.uuid].close()
            del self.__data_group_item_removed_listeners[object.uuid]
            self.__data_group_property_changed_listeners[object.uuid].close()
            del self.__data_group_property_changed_listeners[object.uuid]
            parent_item.remove_child(parent_item.children[index])
            self.__mapping.pop(object)
            self.item_model_controller.end_remove()
            if callable(self.on_item_count_changed):
                self.on_item_count_changed()

    def __update_item_count(self, data_group):
        assert isinstance(data_group, DataGroup.DataGroup)
        count = self.__get_display_item_count_flat(data_group)
        item = self.__mapping[data_group]
        item.data["display"] = str(data_group) + (" (%i)" % count)
        item.data["edit"] = data_group.title
        self.item_model_controller.data_changed(item.row, item.parent.row, item.parent.id)

    # this method if called when one of our listened to data groups changes
    def __display_item_inserted(self, container, display_item, before_index, moving):
        self.__update_item_count(container)

    # this method if called when one of our listened to data groups changes
    def __display_item_removed(self, container, display_item, index, moving):
        self.__update_item_count(container)

    def item_set_data(self, data, index, parent_row, parent_id):
        data_group = self.item_model_controller.item_value("data_group", index, parent_id)
        if data_group:
            command = self.__document_controller.create_rename_data_group_command(data_group, title=data)
            command.perform()
            self.__document_controller.push_undo_command(command)
            return True
        return False

    def get_data_group(self, index, parent_row, parent_id):
        return self.item_model_controller.item_value("data_group", index, parent_id)

    def get_data_group_of_parent(self, parent_row, parent_id):
        parent_item = self.item_model_controller.item_from_id(parent_id)
        return parent_item.data["data_group"] if "data_group" in parent_item.data else None

    def get_data_group_index(self, data_group):
        item = None
        data_group_item = self.__mapping.get(data_group)
        parent_item = data_group_item.parent if data_group_item else self.item_model_controller.root
        assert parent_item is not None
        for child in parent_item.children:
            child_data_group = child.data.get("data_group")
            if child_data_group == data_group:
                item = child
                break
        if item:
            return item.row, item.parent.row, item.parent.id
        else:
            return -1, -1, 0

    def can_drop_mime_data(self, mime_data, action, row, parent_row, parent_id):
        data_group = self.get_data_group_of_parent(parent_row, parent_id)
        if data_group and mime_data.has_file_paths:
            return row < 0  # only accept drops ONTO items, not BETWEEN items
        if data_group and (mime_data.has_format(MimeTypes.DISPLAY_ITEM_MIME_TYPE)):
            if row >= 0:  # only accept drops ONTO items, not BETWEEN items
                return False
            # if the display item exists in this document, then it is copied to the
            # target group. if it doesn't exist in this document, then it is coming
            # from another document and can't be handled here.
            display_item_uuid = uuid.UUID(mime_data.data_as_string(MimeTypes.DISPLAY_ITEM_MIME_TYPE))
            display_item = self.__document_model.get_display_item_by_uuid(display_item_uuid)
            if display_item:
                return True
            return False
        if mime_data.has_format(MimeTypes.DATA_GROUP_MIME_TYPE):
            data_group_uuid = uuid.UUID(mime_data.data_as_string(MimeTypes.DATA_GROUP_MIME_TYPE))
            data_group = self.__document_model.get_data_group_by_uuid(data_group_uuid)
            if data_group:
                return True
        return False

    def item_drop_mime_data(self, mime_data, action, row, parent_row, parent_id):
        data_group = self.get_data_group_of_parent(parent_row, parent_id)
        container = self.__document_model if parent_row < 0 and parent_id == 0 else data_group
        if data_group and mime_data.has_file_paths:
            if row >= 0:  # only accept drops ONTO items, not BETWEEN items
                return self.item_model_controller.NONE
            if callable(self.on_receive_files):
                if self.on_receive_files(mime_data.file_paths, data_group, len(data_group.display_items)):
                    return self.item_model_controller.COPY
        if data_group and (mime_data.has_format(MimeTypes.DISPLAY_ITEM_MIME_TYPE)):
            if row >= 0:  # only accept drops ONTO items, not BETWEEN items
                return self.item_model_controller.NONE
            # if the display item exists in this document, then it is copied to the
            # target group. if it doesn't exist in this document, then it is coming
            # from another document and can't be handled here.
            display_item_uuid = uuid.UUID(mime_data.data_as_string(MimeTypes.DISPLAY_ITEM_MIME_TYPE))
            display_item = self.__document_model.get_display_item_by_uuid(display_item_uuid)
            if display_item:
                command = self.__document_controller.create_insert_data_group_display_item_command(data_group, len(data_group.display_items), display_item)
                command.perform()
                self.__document_controller.push_undo_command(command)
                return action
            return self.item_model_controller.NONE
        if mime_data.has_format(MimeTypes.DATA_GROUP_MIME_TYPE):
            data_group_uuid = uuid.UUID(mime_data.data_as_string(MimeTypes.DATA_GROUP_MIME_TYPE))
            data_group = self.__document_model.get_data_group_by_uuid(data_group_uuid)
            if data_group:
                data_group_copy = copy.deepcopy(data_group)
                row = row if row >= 0 else len(container.data_groups)
                command = self.__document_controller.create_insert_data_group_command(container, row, data_group_copy)
                command.perform()
                self.__document_controller.push_undo_command(command)
                return action
        return self.item_model_controller.NONE

    def item_mime_data(self, index, parent_row, parent_id):
        data_group = self.get_data_group(index, parent_row, parent_id)
        if data_group:
            mime_data = self.ui.create_mime_data()
            mime_data.set_data_as_string(MimeTypes.DATA_GROUP_MIME_TYPE, str(data_group.uuid))
            return mime_data
        return None

    def remove_rows(self, row, count, parent_row, parent_id):
        data_group = self.get_data_group_of_parent(parent_row, parent_id)
        container = self.__document_model if parent_row < 0 and parent_id == 0 else data_group
        for i in range(count):
            del container.data_groups[row]
        return True


class DataPanel(Panel.Panel):

    def __init__(self, document_controller, panel_id, properties):
        super().__init__(document_controller, panel_id, _("Data Items"))

        ui = document_controller.ui

        self.__filter_changed_event_listener = document_controller.filter_changed_event.listen(self.__data_panel_filter_changed)

        class DisplayItemController:

            def __init__(self, base_title, display_items_model):
                self.__base_title = base_title
                self.__count = 0
                self.__display_items_model = display_items_model
                self.on_title_changed = None

                # not thread safe. must be called on ui thread.
                def display_item_inserted(key, display_item, before_index):
                    self.__count += 1
                    if self.on_title_changed:
                        document_controller.queue_task(functools.partial(self.on_title_changed, self.title))

                # not thread safe. must be called on ui thread.
                def display_item_removed(key, display_item, index):
                    self.__count -= 1
                    if self.on_title_changed:
                        document_controller.queue_task(functools.partial(self.on_title_changed, self.title))

                self.__display_item_inserted_listener = self.__display_items_model.item_inserted_event.listen(display_item_inserted)
                self.__display_item_removed_listener = self.__display_items_model.item_removed_event.listen(display_item_removed)

                self.__count = len(self.__display_items_model.display_items)

            @property
            def title(self):
                return self.__base_title + (" (%i)" % self.__count)

            def close(self):
                self.__display_item_inserted_listener.close()
                self.__display_item_inserted_listener = None
                self.__display_item_removed_listener.close()
                self.__display_item_removed_listener = None
                self.__display_items_model.close()

        all_items_controller = DisplayItemController(_("All"), document_controller.create_display_items_model(None, "all"))
        live_items_controller = DisplayItemController(_("Live"), document_controller.create_display_items_model(None, "temporary"))
        latest_items_controller = DisplayItemController(_("Latest Session"), document_controller.create_display_items_model(None, "latest-session"))
        self.__item_controllers = [all_items_controller, live_items_controller, latest_items_controller]

        self.library_model_controller = LibraryModelController(ui, self.__item_controllers)
        self.library_model_controller.on_receive_files = self.library_model_receive_files

        self.data_group_model_controller = DataGroupModelController(ui, document_controller)
        self.data_group_model_controller.on_receive_files = lambda file_paths, data_group, index: self.data_group_model_receive_files(file_paths, data_group, index)

        self.__blocked1 = False
        self.__blocked2 = False

        def library_widget_selection_changed(selected_indexes):
            # this message comes from the collection tree widget when the selection changes
            if not self.__blocked1:
                self.__blocked1 = True
                try:
                    index = selected_indexes[0][0] if len(selected_indexes) > 0 else -1
                    if index == 2:
                        document_controller.set_filter("latest-session")
                    elif index == 1:
                        document_controller.set_filter("temporary")
                    else:
                        document_controller.set_filter("all")
                finally:
                    self.__blocked1 = False

        self.library_widget = ui.create_tree_widget()
        self.library_widget.item_model_controller = self.library_model_controller.item_model_controller
        self.library_widget.on_selection_changed = library_widget_selection_changed
        self.library_widget.on_focus_changed = lambda focused: setattr(self, "focused", focused)

        self.library_model_controller.on_item_count_changed = self.library_widget.size_to_content
        self.library_widget.size_to_content()

        def data_group_widget_selection_changed(selected_indexes):
            # this message comes from the data group tree widget when the selection changes
            if not self.__blocked2:
                self.__blocked2 = True
                try:
                    if len(selected_indexes) > 0:
                        index, parent_row, parent_id = selected_indexes[0]
                        data_group = self.data_group_model_controller.get_data_group(index, parent_row, parent_id)
                    else:
                        data_group = None
                    document_controller.set_data_group(data_group)
                finally:
                    self.__blocked2 = False

        def data_group_widget_key_pressed(index, parent_row, parent_id, key):
            if key.is_delete:
                data_group = self.data_group_model_controller.get_data_group(index, parent_row, parent_id)
                if data_group:
                    container = self.data_group_model_controller.get_data_group_of_parent(parent_row, parent_id)
                    container = container if container else self.document_controller.document_model
                    self.document_controller.remove_data_group_from_container(data_group, container)
                return True
            return False

        self.data_group_widget = ui.create_tree_widget()
        self.data_group_widget.item_model_controller = self.data_group_model_controller.item_model_controller
        self.data_group_widget.on_selection_changed = data_group_widget_selection_changed
        self.data_group_widget.on_item_key_pressed = data_group_widget_key_pressed
        self.data_group_widget.on_focus_changed = lambda focused: setattr(self, "focused", focused)

        self.data_group_model_controller.on_item_count_changed = self.data_group_widget.size_to_content
        self.data_group_widget.size_to_content()

        library_label_row = ui.create_row_widget()
        library_label_row.add_spacing(8)
        library_label_row.add(ui.create_label_widget(_("Library"), properties={"stylesheet": "font-weight: bold"}))
        library_label_row.add_stretch()

        collections_label_row = ui.create_row_widget()
        collections_label_row.add_spacing(8)
        collections_label_row.add(ui.create_label_widget(_("Collections"), properties={"stylesheet": "font-weight: bold"}))
        collections_label_row.add_stretch()

        library_section_widget = ui.create_column_widget()
        library_section_widget.add_spacing(4)
        library_section_widget.add(library_label_row)
        library_section_widget.add(self.library_widget)
        library_section_widget.add_spacing(4)
        library_section_widget.add(collections_label_row)
        library_section_widget.add(self.data_group_widget)
        library_section_widget.add_spacing(4)
        library_section_widget.add_stretch()

        def show_context_menu(display_item_adapter, x, y, gx, gy):
            menu = document_controller.create_context_menu_for_display(display_item_adapter.display_item, use_selection=True)
            menu.popup(gx, gy)
            return True

        def map_display_item_to_display_item_adapter(display_item):
            return DisplayItemAdapter(display_item, ui)

        def unmap_display_item_to_display_item_adapter(display_item_adapter):
            display_item_adapter.close()

        self.__filtered_display_item_adapters_model = ListModel.MappedListModel(container=document_controller.filtered_display_items_model, master_items_key="display_items", items_key="display_item_adapters", map_fn=map_display_item_to_display_item_adapter, unmap_fn=unmap_display_item_to_display_item_adapter)

        self.__selection = self.document_controller.selection

        self.__focused = False

        def selection_changed():
            # called when the selection changes; notify selected display item changed if focused.
            self.__notify_focus_changed()

        self.__selection_changed_event_listener = self.__selection.changed_event.listen(selection_changed)

        def display_item_adapter_selection_changed(display_item_adapters):
            indexes = set()
            for index, display_item_adapter in enumerate(self.__filtered_display_item_adapters_model.display_item_adapters):
                if display_item_adapter in display_item_adapters:
                    indexes.add(index)
            self.__selection.set_multiple(indexes)
            self.__notify_focus_changed()

        def focus_changed(focused):
            self.focused = focused

        def delete_display_item_adapters(display_item_adapters):
            document_controller.delete_display_items([display_item_adapter.display_item for display_item_adapter in display_item_adapters])

        self.data_list_controller = DataListController(document_controller.event_loop, ui, self.__filtered_display_item_adapters_model, self.__selection)
        self.data_list_controller.on_display_item_adapter_selection_changed = display_item_adapter_selection_changed
        self.data_list_controller.on_context_menu_event = show_context_menu
        self.data_list_controller.on_focus_changed = focus_changed
        self.data_list_controller.on_delete_display_item_adapters = delete_display_item_adapters

        self.data_grid_controller = DataGridController(document_controller.event_loop, ui, self.__filtered_display_item_adapters_model, self.__selection)
        self.data_grid_controller.on_display_item_adapter_selection_changed = display_item_adapter_selection_changed
        self.data_grid_controller.on_context_menu_event = show_context_menu
        self.data_grid_controller.on_focus_changed = focus_changed
        self.data_grid_controller.on_delete_display_item_adapters = delete_display_item_adapters

        data_list_widget = DataListWidget(ui, self.data_list_controller)
        data_grid_widget = DataGridWidget(ui, self.data_grid_controller)

        list_icon_button = CanvasItem.BitmapButtonCanvasItem(CanvasItem.load_rgba_data_from_bytes(pkgutil.get_data(__name__, "resources/list_icon_20.png")))
        grid_icon_button = CanvasItem.BitmapButtonCanvasItem(CanvasItem.load_rgba_data_from_bytes(pkgutil.get_data(__name__, "resources/grid_icon_20.png")))

        list_icon_button.sizing.set_fixed_size(Geometry.IntSize(20, 20))
        grid_icon_button.sizing.set_fixed_size(Geometry.IntSize(20, 20))

        button_row = CanvasItem.CanvasItemComposition()
        button_row.layout = CanvasItem.CanvasItemRowLayout(spacing=4)
        button_row.add_canvas_item(list_icon_button)
        button_row.add_canvas_item(grid_icon_button)

        buttons_widget = ui.create_canvas_widget(properties={"height": 20, "width": 44})
        buttons_widget.canvas_item.add_canvas_item(button_row)

        search_widget = ui.create_row_widget()
        search_widget.add_spacing(8)
        search_widget.add(ui.create_label_widget(_("Filter")))
        search_widget.add_spacing(8)
        search_line_edit = ui.create_line_edit_widget()
        search_line_edit.placeholder_text = _("No Filter")
        # search_line_edit.clear_button_enabled = True  # Qt 5.3 doesn't signal text edited or editing finished when clearing. useless so disabled.
        search_line_edit.on_text_edited = self.document_controller.filter_controller.text_filter_changed
        search_line_edit.on_editing_finished = self.document_controller.filter_controller.text_filter_changed
        search_widget.add(search_line_edit)
        search_widget.add_spacing(6)
        search_widget.add(buttons_widget)
        search_widget.add_spacing(8)

        self.data_view_widget = ui.create_stack_widget()
        self.data_view_widget.add(data_list_widget)
        self.data_view_widget.add(data_grid_widget)
        self.data_view_widget.current_index = 0

        self.__view_button_group = CanvasItem.RadioButtonGroup([list_icon_button, grid_icon_button])
        self.__view_button_group.current_index = 0
        self.__view_button_group.on_current_index_changed = lambda index: setattr(self.data_view_widget, "current_index", index)

        slave_widget = ui.create_column_widget()
        slave_widget.add(self.data_view_widget)
        slave_widget.add_spacing(6)
        slave_widget.add(search_widget)
        slave_widget.add_spacing(6)

        self.splitter = ui.create_splitter_widget("vertical", properties)
        self.splitter.orientation = "vertical"
        self.splitter.add(library_section_widget)
        self.splitter.add(slave_widget)
        self.splitter.set_sizes([1, 9999])  # minimum library section; maximum display item section
        self.splitter.restore_state("window/v1/data_panel_splitter")

        self.widget = self.splitter

        self._data_list_widget = data_list_widget
        self._data_grid_widget = data_grid_widget

        data_group, filter_id = document_controller.get_data_group_and_filter_id()
        self.__data_panel_filter_changed(data_group, filter_id)

    def close(self):
        # display items model should not be closed since it isn't created in this object
        self.splitter.save_state("window/v1/data_panel_splitter")
        # close the widget to stop repainting the widgets before closing the controllers.
        super().close()
        # finish closing
        self.data_list_controller.close()
        self.data_list_controller = None
        self.data_grid_controller.close()
        self.data_grid_controller = None
        # close the item models
        self.data_group_model_controller.close()
        self.data_group_model_controller = None
        for item_controller in self.__item_controllers:
            item_controller.close()
        self.library_model_controller.close()
        self.library_model_controller = None
        # and the listeners
        self.__filter_changed_event_listener.close()
        self.__filter_changed_event_listener = None
        self.__filtered_display_item_adapters_model.close()
        self.__filtered_display_item_adapters_model = None
        # button group
        self.__view_button_group.close()
        self.__view_button_group = None

    def __notify_focus_changed(self):
        if self.__focused:
            if len(self.__selection.indexes) == 1:
                display_item_adapter = self.__filtered_display_item_adapters_model.display_item_adapters[list(self.__selection.indexes)[0]]
                self.document_controller.notify_focused_display_changed(display_item_adapter.display_item)
            else:
                self.document_controller.notify_focused_display_changed(None)

    @property
    def focused(self):
        return self.__focused

    @focused.setter
    def focused(self, value):
        self.__focused = value
        self.__notify_focus_changed()

    def __data_panel_filter_changed(self, data_group, filter_id):
        if data_group:
            index, parent_row, parent_id = self.data_group_model_controller.get_data_group_index(data_group)
            self.library_widget.clear_current_row()
            self.data_group_widget.set_current_row(index, parent_row, parent_id)
        else:
            self.data_group_widget.clear_current_row()
            if filter_id == "latest-session":
                self.library_widget.set_current_row(2, -1, 0)  # select the 'latest' group
            elif filter_id == "temporary":
                self.library_widget.set_current_row(1, -1, 0)  # select the 'live' group
            else:
                self.library_widget.set_current_row(0, -1, 0)  # select the 'all' group

    def library_model_receive_files(self, file_paths, threaded=True):
        self.document_controller._register_ui_activity()

        def receive_files_complete(received_data_items):
            def select_library_all():
                self.document_controller.select_data_items_in_data_panel([received_data_items[0]])
            if len(received_data_items) > 0:
                self.queue_task(select_library_all)

        self.document_controller.receive_files(file_paths, None, -1, threaded, receive_files_complete)
        return True

    # receive files dropped into the data group.
    # this message comes from the data group model, which is why it is named the way it is.
    def data_group_model_receive_files(self, file_paths, data_group, index, threaded=True):
        def receive_files_complete(received_data_items):
            def select_data_group_and_data_item():
                self.document_controller.set_data_group(data_group)
                self.document_controller.select_data_items_in_data_panel([received_data_items[0]])
            if len(received_data_items) > 0:
                if threaded:
                    self.queue_task(select_data_group_and_data_item)
                else:
                    select_data_group_and_data_item()

        self.document_controller.receive_files(file_paths, data_group, index, threaded, receive_files_complete)
        return True



# standard libraries
import gettext
import typing

# third party libraries
# None

# local libraries
from nion.swift import Panel

_ = gettext.gettext


class InfoPanel(Panel.Panel):
    """A panel to display cursor information.

    The info panel will display cursor information. User interface items that want to
    update the cursor info should called cursor_changed on the document controller.
    This info panel will listen to the document controller for cursor updates and update
    itself in response. all cursor update calls are thread safe. this class uses periodic
    to do ui updates from the main thread.
    """

    def __init__(self, document_controller, panel_id, properties):
        super(InfoPanel, self).__init__(document_controller, panel_id, _("Info"))

        ui = document_controller.ui

        self.label_row_1 = ui.create_label_widget()
        self.label_row_2 = ui.create_label_widget()
        self.label_row_3 = ui.create_label_widget()

        text_row_1 = ui.create_row_widget(properties={"spacing": 6})
        text_row_1.add(self.label_row_1)
        text_row_1.add_stretch()

        text_row_2 = ui.create_row_widget(properties={"spacing": 6})
        text_row_2.add(self.label_row_2)
        text_row_2.add_stretch()

        text_row_3 = ui.create_row_widget(properties={"spacing": 6})
        text_row_3.add(self.label_row_3)
        text_row_3.add_stretch()

        properties["spacing"] = 2
        properties["margin"] = 6
        column = ui.create_column_widget(properties=properties)
        column.add(text_row_1)
        column.add(text_row_2)
        column.add(text_row_3)
        column.add_stretch()

        self.widget = column

        # connect self as listener. this will result in calls to cursor_changed
        self.__cursor_changed_event_listener = self.document_controller.cursor_changed_event.listen(self.__cursor_changed)

    def close(self):
        # disconnect self as listener
        self.__cursor_changed_event_listener.close()
        self.__cursor_changed_event_listener = None
        self.clear_task("position_and_value")
        # finish closing
        super(InfoPanel, self).close()

    # this message is received from the document controller.
    def __cursor_changed(self, text_items: typing.List[str]) -> None:
        def update_position_and_value(text_items: typing.List[str]):
            self.label_row_1.text = text_items[0] if text_items and len(text_items) > 0 else None
            self.label_row_2.text = text_items[1] if text_items and len(text_items) > 1 else None
            self.label_row_3.text = text_items[2] if text_items and len(text_items) > 2 else None

        self.add_task("position_and_value", lambda: update_position_and_value(text_items))



# standard libraries
import importlib
import inspect
import os
import sys
import unittest

# third party libraries
# None

# local libraries
from nion.swift.model import PlugInManager


suites = []
suite_dict = {}
alltests = None


# scan through directory and look for tests (files ending in test.py)
# load the module and add the tests
def load_tests(packages):
    global suites
    global suite_dict
    global alltests

    base_prefix = sys.base_prefix
    directories = set()
    for path in sys.path:
        if not path.startswith(base_prefix):
            directories.add(path)
    for package in packages:
        for directory in directories:
            test_dir = os.path.join(directory, *package.split("."), "test")
            if os.path.isdir(test_dir):
                for file in os.listdir(test_dir):
                    if file.endswith("_test.py"):
                        module_name = package + ".test." + file.replace('.py', '')
                        module = importlib.import_module(module_name)
                        for maybe_a_class in inspect.getmembers(module):
                            if inspect.isclass(maybe_a_class[1]) and maybe_a_class[0].startswith("Test"):
                                test_name = maybe_a_class[0]
                                # It is a class... add it to the test suite.
                                cls = getattr(module, test_name)
                                suite = unittest.TestLoader().loadTestsFromTestCase(cls)
                                suites.append(suite)
                                suite_dict[test_name] = suite

    suites.extend(PlugInManager.test_suites())

    alltests = unittest.TestSuite(suites)

def run_all_tests():
    unittest.TextTestRunner(verbosity=2).run(alltests)

def run_test(test_name):
    unittest.TextTestRunner(verbosity=2).run(suite_dict[test_name])



# system imports
import code
import contextlib
import copy
import gettext
import io
import re
import rlcompleter
import sys

# local libraries
from nion.ui import Dialog
from nion.ui import Widgets

_ = gettext.gettext


@contextlib.contextmanager
def reassign_stdout(new_stdout, new_stderr):
    oldstdout, oldtsderr = sys.stdout, sys.stderr
    sys.stdout, sys.stderr = new_stdout, new_stderr
    yield
    sys.stdout, sys.stderr = oldstdout, oldtsderr


def get_common_prefix(l):
    if not l:
        return str()
    s1 = min(l)  # return the first, alphabetically
    s2 = max(l)  # the last
    # check common characters in between
    for i, c in enumerate(s1):
        if c != s2[i]:
            return s1[:i]
    return s1


class ConsoleWidget(Widgets.CompositeWidgetBase):

    def __init__(self, ui, locals=None, properties=None):
        super().__init__(ui.create_column_widget())

        properties = properties if properties is not None else dict()
        properties["stylesheet"] = "background: black; color: white; font-family: Monaco, Courier, monospace"

        self.prompt = ">>> "
        self.continuation_prompt = "... "

        self.__cursor_position = None

        self.__text_edit_widget = ui.create_text_edit_widget(properties)
        self.__text_edit_widget.word_wrap_mode = "anywhere"
        self.__text_edit_widget.on_cursor_position_changed = self.__cursor_position_changed
        self.__text_edit_widget.on_selection_changed = self.__selection_changed
        self.__text_edit_widget.on_return_pressed = self.__return_pressed
        self.__text_edit_widget.on_key_pressed = self.__key_pressed
        self.__text_edit_widget.on_insert_mime_data = self.__insert_mime_data

        class StdoutCatcher:
            def __init__(self, text_edit_widget):
                self.__text_edit_widget = text_edit_widget
            def write(self, stuff):
                if self.__text_edit_widget:
                    self.__text_edit_widget.append_text(str(stuff).rstrip())
            def flush(self):
                pass
        stdout = StdoutCatcher(self.__text_edit_widget)

        locals = locals if locals is not None else dict()
        locals.update({'__name__': None, '__console__': None, '__doc__': None, '_stdout': stdout})
        self.console = code.InteractiveConsole(locals)
        self.__text_edit_widget.append_text(self.prompt)
        self.__text_edit_widget.move_cursor_position("end")
        self.__last_position = copy.deepcopy(self.__cursor_position)

        self.content_widget.add(self.__text_edit_widget)

        self.__history = list()
        self.__history_point = None

    def close(self):
        super().close()
        self.__text_edit_widget = None

    def insert_lines(self, lines):
        for l in lines:
            self.__text_edit_widget.move_cursor_position("end")
            self.__text_edit_widget.insert_text(l)
            result, error_code, prompt = self.interpret_command(l)
            if len(result) > 0:
                self.__text_edit_widget.set_text_color("red" if error_code else "green")
                self.__text_edit_widget.append_text(result[:-1])
                self.__text_edit_widget.set_text_color("white")
            self.__text_edit_widget.append_text(prompt)
            self.__text_edit_widget.move_cursor_position("end")
            self.__last_position = copy.deepcopy(self.__cursor_position)
            if l: self.__history.append(l)
            self.__history_point = None

    # interpretCommand is called from the intrinsic widget.
    def interpret_command(self, command):
        output = io.StringIO()
        error = io.StringIO()
        with reassign_stdout(output, error):
            self.__incomplete = self.console.push(command)
        prompt = self.continuation_prompt if self.__incomplete else self.prompt
        if error.getvalue():
            result =  error.getvalue()
            error_code = -1
        else:
            result = output.getvalue()
            error_code = 0
        return result, error_code, prompt

    def interpret_lines(self, lines):
        for l in lines:
            self.interpret_command(l)

    def __return_pressed(self):
        command = self.__get_partial_command()
        result, error_code, prompt = self.interpret_command(command)
        if len(result) > 0:
            self.__text_edit_widget.set_text_color("red" if error_code else "green")
            self.__text_edit_widget.append_text(result[:-1])
            self.__text_edit_widget.set_text_color("white")
        self.__text_edit_widget.append_text(prompt)
        self.__text_edit_widget.move_cursor_position("end")
        self.__last_position = copy.deepcopy(self.__cursor_position)
        if command: self.__history.append(command)
        self.__history_point = None
        return True

    def __get_partial_command(self):
        command = self.__text_edit_widget.text.split('\n')[-1]
        if command.startswith(self.prompt):
            command = command[len(self.prompt):]
        elif command.startswith(self.continuation_prompt):
            command = command[len(self.continuation_prompt):]
        return command

    def __key_pressed(self, key):
        is_cursor_on_last_line = self.__cursor_position.block_number == self.__last_position.block_number

        if is_cursor_on_last_line and key.is_up_arrow:
            if self.__history_point is None:
                self.__history_point = len(self.__history)
            self.__history_point = max(0, self.__history_point - 1)
            if self.__history_point < len(self.__history):
                line = self.__history[self.__history_point]
                self.__text_edit_widget.move_cursor_position("start_para", "move")
                self.__text_edit_widget.move_cursor_position("end_para", "keep")
                prompt = self.continuation_prompt if self.__incomplete else self.prompt
                self.__text_edit_widget.insert_text("{}{}".format(prompt, line))
                self.__text_edit_widget.move_cursor_position("end")
                self.__last_position = copy.deepcopy(self.__cursor_position)
            return True

        if is_cursor_on_last_line and key.is_down_arrow:
            if self.__history_point is not None:
                self.__history_point = min(len(self.__history), self.__history_point + 1)
                if self.__history_point < len(self.__history):
                    line = self.__history[self.__history_point]
                else:
                    self.__history_point = None
                    line = ""
                self.__text_edit_widget.move_cursor_position("start_para", "move")
                self.__text_edit_widget.move_cursor_position("end_para", "keep")
                prompt = self.continuation_prompt if self.__incomplete else self.prompt
                self.__text_edit_widget.insert_text("{}{}".format(prompt, line))
                self.__text_edit_widget.move_cursor_position("end")
                self.__last_position = copy.deepcopy(self.__cursor_position)

        if is_cursor_on_last_line and key.is_delete:
            partial_command = self.__get_partial_command()
            if not partial_command:
                return True

        if is_cursor_on_last_line and key.is_move_to_start_of_line:
            mode = "keep" if key.modifiers.shift else "move"
            self.__text_edit_widget.move_cursor_position("start_para", mode)
            self.__text_edit_widget.move_cursor_position("next", mode, n=4)
            return True

        if is_cursor_on_last_line and key.is_delete_to_end_of_line:
            self.__text_edit_widget.move_cursor_position("end_para", "keep")
            self.__text_edit_widget.remove_selected_text()
            return True

        if is_cursor_on_last_line and key.key == 0x43 and key.modifiers.native_control and sys.platform == "darwin":
            prompt = self.continuation_prompt if self.__incomplete else self.prompt
            self.__text_edit_widget.move_cursor_position("end")
            self.__text_edit_widget.insert_text("\n")
            self.__text_edit_widget.insert_text(prompt)
            self.__text_edit_widget.move_cursor_position("end")
            self.__last_position = copy.deepcopy(self.__cursor_position)
            return True

        if is_cursor_on_last_line and key.is_tab:
            partial_command = self.__get_partial_command()
            terms = list()
            completer = rlcompleter.Completer(namespace=self.console.locals)
            index = 0
            delims = " \t\n`~!@#$%^&*()-=+[{]}\\|;:\'\",<>/?"
            rx = "[" + re.escape("".join(delims)) + "]"
            split_commands = re.split(rx, partial_command)
            if len(split_commands) > 0:
                completion_term = split_commands[-1]
                while True:
                    term = completer.complete(completion_term, index)
                    if term is None:
                        break
                    index += 1
                    if not term.startswith(completion_term + "__"):
                        terms.append(term)
                if len(terms) == 1:
                    completed_command = partial_command[:partial_command.rfind(completion_term)] + terms[0]
                    prompt = self.continuation_prompt if self.__incomplete else self.prompt
                    self.__text_edit_widget.move_cursor_position("start_para", "move")
                    self.__text_edit_widget.move_cursor_position("end_para", "keep")
                    self.__text_edit_widget.insert_text("{}{}".format(prompt, completed_command))
                    self.__text_edit_widget.move_cursor_position("end")
                    self.__last_position = copy.deepcopy(self.__cursor_position)
                elif len(terms) > 1:
                    common_prefix = get_common_prefix(terms)
                    prompt = self.continuation_prompt if self.__incomplete else self.prompt
                    self.__text_edit_widget.move_cursor_position("end")
                    self.__text_edit_widget.set_text_color("brown")
                    self.__text_edit_widget.append_text("   ".join(terms) + "\n")
                    self.__text_edit_widget.move_cursor_position("end")
                    self.__text_edit_widget.set_text_color("white")
                    self.__text_edit_widget.insert_text("{}{}".format(prompt, common_prefix))
                    self.__text_edit_widget.move_cursor_position("end")
                    self.__last_position = copy.deepcopy(self.__cursor_position)
                return True

        return False

    def __cursor_position_changed(self, cursor_position):
        self.__cursor_position = copy.deepcopy(cursor_position)

    def __selection_changed(self, selection):
        pass

    def __insert_mime_data(self, mime_data):
        text = mime_data.data_as_string("text/plain")
        text_lines = re.split("[" + re.escape("\n") + re.escape("\r") + "]", text)
        if text_lines[-1] == "":
            text_lines = text_lines[:-1]
        if len(text_lines) == 1 and text_lines[0] == text.rstrip():
            # special case where text has no line terminator
            self.__text_edit_widget.insert_text(text)
        else:
            self.insert_lines(text_lines)



class ConsoleDialog(Dialog.ActionDialog):

    def __init__(self, document_controller):
        super().__init__(document_controller.ui, _("Python Console"), document_controller.app, persistent_id="ConsoleDialog")

        self.__document_controller = document_controller

        self.__console_widget = ConsoleWidget(document_controller.ui, properties={"min-height": 180, "min-width": 540})

        lines = [
            "import logging",
            "import numpy as np",
            "import numpy as numpy",
            "import uuid",
            "from nion.swift.model import PlugInManager",
            "from nion.ui import Declarative",
            "from nion.data import xdata_1_0 as xd",
            "get_api = PlugInManager.api_broker_fn",
            "api = get_api('~1.0', '~1.0')",
            "ui = Declarative.DeclarativeUI()",
            "show = api.show",
            "def run_script(*args, **kwargs):",
            "  api.run_script(*args, stdout=_stdout, **kwargs)",
            "",
            ]

        variable_to_data_item_map = document_controller.document_model.variable_to_data_item_map()
        for variable_name, data_item in variable_to_data_item_map.items():
            lines.append("{} = api.library.get_data_item_by_uuid(uuid.UUID(\"{}\"))".format(variable_name, data_item.uuid))

        self.__console_widget.interpret_lines(lines)

        self.content.add(self.__console_widget)

        self.__document_controller.register_console(self)

        self._create_menus()

    def close(self):
        self.__document_controller.unregister_console(self)
        super().close()

    def assign_data_item_var(self, data_item_var, data_item):
        self.__console_widget.insert_lines(["{} = api.library.get_data_item_by_uuid(uuid.UUID(\"{}\"))".format(data_item_var, data_item.uuid)])



# standard libraries
# None

# third party libraries
# None

# local libraries
from nion.data import Image
from nion.swift import MimeTypes
from nion.swift import Thumbnails
from nion.swift.model import DisplayItem
from nion.swift.model import DocumentModel
from nion.ui import CanvasItem
from nion.ui import Widgets
from nion.utils import Geometry


class AbstractThumbnailSource:

    def __init__(self):
        self.on_thumbnail_data_changed = None
        self.__thumbnail_data = None
        self.overlay_canvas_item = CanvasItem.EmptyCanvasItem()

    def close(self):
        self.on_thumbnail_data_changed = None

    @property
    def thumbnail_data(self):
        return self.__thumbnail_data

    def _set_thumbnail_data(self, thumbnail_data):
        self.__thumbnail_data = thumbnail_data

    def populate_mime_data_for_drag(self, mime_data, size: Geometry.IntSize):
        return False, None


class BitmapOverlayCanvasItem(CanvasItem.CanvasItemComposition):

    def __init__(self):
        super().__init__()
        self.focusable = True
        self.__dropping = False
        self.__focused = False
        self.wants_drag_events = True
        self.wants_mouse_events = True
        self.__drag_start = None
        self.on_drop_mime_data = None
        self.on_delete = None
        self.on_drag_pressed = None
        self.active = False

    def close(self):
        self.on_drop_mime_data = None
        self.on_delete = None
        self.on_drag_pressed = None
        super().close()

    @property
    def focused(self):
        return self.__focused

    def _set_focused(self, focused):
        if self.__focused != focused:
            self.__focused = focused
            self.update()

    def _repaint(self, drawing_context):
        super()._repaint(drawing_context)
        # canvas size
        canvas_width = self.canvas_size[1]
        canvas_height = self.canvas_size[0]
        focused_style = "#3876D6"  # TODO: platform dependent
        if self.active:
            with drawing_context.saver():
                drawing_context.begin_path()
                drawing_context.round_rect(2, 2, 6, 6, 3)
                drawing_context.fill_style = "rgba(0, 255, 0, 0.80)"
                drawing_context.fill()
        if self.__dropping:
            with drawing_context.saver():
                drawing_context.begin_path()
                drawing_context.rect(0, 0, canvas_width, canvas_height)
                drawing_context.fill_style = "rgba(255, 0, 0, 0.10)"
                drawing_context.fill()
        if self.focused:
            stroke_style = focused_style
            drawing_context.begin_path()
            drawing_context.rect(2, 2, canvas_width - 4, canvas_height - 4)
            drawing_context.line_join = "miter"
            drawing_context.stroke_style = stroke_style
            drawing_context.line_width = 4.0
            drawing_context.stroke()

    def drag_enter(self, mime_data):
        self.__dropping = True
        self.update()
        return "ignore"

    def drag_leave(self):
        self.__dropping = False
        self.update()
        return False

    def drop(self, mime_data, x, y):
        if callable(self.on_drop_mime_data):
            result = self.on_drop_mime_data(mime_data, x, y)
            if result:
                return result
        return super().drop(mime_data, x, y)

    def key_pressed(self, key):
        if key.is_delete:
            on_delete = self.on_delete
            if callable(on_delete):
                on_delete()
                return True
        return super().key_pressed(key)

    def mouse_pressed(self, x, y, modifiers):
        self.__drag_start = Geometry.IntPoint(x=x, y=y)
        return True

    def mouse_released(self, x, y, modifiers):
        self.__drag_start = None
        return True

    def mouse_position_changed(self, x, y, modifiers):
        p = Geometry.IntPoint(x=x, y=y)
        if self.__drag_start is not None and Geometry.distance(p, self.__drag_start) > 2:
            self.__drag_start = None
            on_drag_pressed = self.on_drag_pressed
            if on_drag_pressed:
                on_drag_pressed(x, y, modifiers)


class ThumbnailCanvasItem(CanvasItem.CanvasItemComposition):

    def __init__(self, ui, thumbnail_source: AbstractThumbnailSource, size: Geometry.IntSize):
        super().__init__()
        bitmap_overlay_canvas_item = BitmapOverlayCanvasItem()
        bitmap_canvas_item = CanvasItem.BitmapCanvasItem(background_color="#CCC", border_color="#444")
        bitmap_canvas_item.sizing.set_fixed_size(size)
        bitmap_overlay_canvas_item.add_canvas_item(bitmap_canvas_item)
        thumbnail_source.overlay_canvas_item.sizing.set_fixed_size(size)
        bitmap_overlay_canvas_item.add_canvas_item(thumbnail_source.overlay_canvas_item)
        self.__thumbnail_source = thumbnail_source
        self.on_drag = None
        self.on_drop_mime_data = None
        self.on_delete = None

        def drag_pressed(x, y, modifiers):
            on_drag = self.on_drag
            if callable(on_drag):
                mime_data = ui.create_mime_data()
                valid, thumbnail = thumbnail_source.populate_mime_data_for_drag(mime_data, size)
                if valid:
                    on_drag(mime_data, thumbnail, x, y)

        def drop_mime_data(mime_data, x, y):
            if callable(self.on_drop_mime_data):
                return self.on_drop_mime_data(mime_data, x, y)
            return None

        def delete():
            on_delete = self.on_delete
            if callable(on_delete):
                on_delete()

        bitmap_overlay_canvas_item.on_drag_pressed = drag_pressed
        bitmap_overlay_canvas_item.on_drop_mime_data = drop_mime_data
        bitmap_overlay_canvas_item.on_delete = delete

        def thumbnail_data_changed(thumbnail_data):
            bitmap_canvas_item.rgba_bitmap_data = thumbnail_data

        self.__thumbnail_source.on_thumbnail_data_changed = thumbnail_data_changed

        bitmap_canvas_item.rgba_bitmap_data = self.__thumbnail_source.thumbnail_data

        self.add_canvas_item(bitmap_overlay_canvas_item)

    def close(self):
        self.__thumbnail_source.close()
        self.__thumbnail_source = None
        self.on_drag = None
        self.on_drop_mime_data = None
        self.on_delete = None
        super().close()


class ThumbnailWidget(Widgets.CompositeWidgetBase):

    def __init__(self, ui, thumbnail_source: AbstractThumbnailSource, size: Geometry.IntSize):
        super().__init__(ui.create_column_widget())
        thumbnail_canvas_item = ThumbnailCanvasItem(ui, thumbnail_source, size)
        bitmap_canvas_widget = ui.create_canvas_widget(properties={"height": size.height, "width": size.width})
        bitmap_canvas_widget.canvas_item.add_canvas_item(thumbnail_canvas_item)
        self.content_widget.add(bitmap_canvas_widget)
        self.on_drop_mime_data = None
        self.on_drag = None
        self.on_delete = None

        def drop_mime_data(mime_data, x, y):
            if callable(self.on_drop_mime_data):
                return self.on_drop_mime_data(mime_data, x, y)
            return None

        def drag(mime_data, thumbnail, x, y):
            on_drag = self.on_drag
            if callable(on_drag):
                on_drag(mime_data, thumbnail, x, y)

        def delete():
            on_delete = self.on_delete
            if callable(on_delete):
                on_delete()

        thumbnail_canvas_item.on_drop_mime_data = drop_mime_data
        thumbnail_canvas_item.on_drag = drag
        thumbnail_canvas_item.on_delete = delete

    def close(self):
        super().close()
        self.on_drop_mime_data = None
        self.on_drag = None
        self.on_delete = None


class DataItemBitmapOverlayCanvasItem(CanvasItem.AbstractCanvasItem):

    def __init__(self):
        super().__init__()
        self.__active = False

    @property
    def active(self):
        return self.__active

    @active.setter
    def active(self, value):
        if value != self.__active:
            self.__active = value
            self.update()

    def _repaint(self, drawing_context):
        super()._repaint(drawing_context)
        if self.active:
            with drawing_context.saver():
                drawing_context.begin_path()
                drawing_context.round_rect(2, 2, 6, 6, 3)
                drawing_context.fill_style = "rgba(0, 255, 0, 0.80)"
                drawing_context.fill()


class DataItemThumbnailSource(AbstractThumbnailSource):

    def __init__(self, ui, *, display_item=None):
        super().__init__()
        self.ui = ui
        self.__display_item = None
        self.__thumbnail_source = None
        self.__thumbnail_updated_event_listener = None
        self.overlay_canvas_item = DataItemBitmapOverlayCanvasItem()
        if display_item:
            self.set_display_item(display_item)

    def close(self):
        self.__detach_listeners()
        super().close()

    def __detach_listeners(self):
        if self.__thumbnail_updated_event_listener:
            self.__thumbnail_updated_event_listener.close()
            self.__thumbnail_updated_event_listener = None
        if self.__thumbnail_source:
            self.__thumbnail_source.close()
            self.__thumbnail_source = None

    def __update_thumbnail(self) -> None:
        if self.__display_item:
            self._set_thumbnail_data(Thumbnails.ThumbnailManager().thumbnail_data_for_display_item(self.__display_item))
            self.overlay_canvas_item.active = self.__display_item.is_live
        else:
            self._set_thumbnail_data(None)
            self.overlay_canvas_item.active = False
        if callable(self.on_thumbnail_data_changed):
            self.on_thumbnail_data_changed(self.thumbnail_data)

    def set_display_item(self, display_item: DisplayItem.DisplayItem) -> None:
        self.__detach_listeners()
        self.__display_item = display_item
        if display_item:
            self.__thumbnail_source = Thumbnails.ThumbnailManager().thumbnail_source_for_display_item(self.ui, display_item)
            self.__thumbnail_updated_event_listener = self.__thumbnail_source.thumbnail_updated_event.listen(self.__update_thumbnail)
        self.__update_thumbnail()

    def populate_mime_data_for_drag(self, mime_data, size: Geometry.IntSize):
        if self.__display_item:
            mime_data.set_data_as_string(MimeTypes.DISPLAY_ITEM_MIME_TYPE, str(self.__display_item.uuid))
            rgba_image_data = self.__thumbnail_source.thumbnail_data
            thumbnail = Image.get_rgba_data_from_rgba(Image.scaled(Image.get_rgba_view_from_rgba_data(rgba_image_data), (size.width, size.height))) if rgba_image_data is not None else None
            return True, thumbnail
        return False, None


class DataItemReferenceThumbnailSource(DataItemThumbnailSource):
    """Used to track a data item referenced by a data item reference.

    Useful, for instance, for displaying a live update thumbnail that can be dragged to other locations."""

    def __init__(self, ui, document_model, data_item_reference: DocumentModel.DocumentModel.DataItemReference):
        display_item = document_model.get_display_item_for_data_item(data_item_reference.data_item)

        super().__init__(ui, display_item=display_item)

        def data_item_changed():
            display_item = document_model.get_display_item_for_data_item(data_item_reference.data_item)
            self.set_display_item(display_item)

        self.__data_item_reference_changed_event_listener = data_item_reference.data_item_reference_changed_event.listen(data_item_changed)

    def close(self):
        self.__data_item_reference_changed_event_listener.close()
        self.__data_item_reference_changed_event_listener = None
        super().close()


DataItemThumbnailWidget = ThumbnailWidget



import threading


all_classes = {
    "API",
    "Application",
    "DataGroup",
    "DataItem",
    "Display",
    "DisplayPanel",
    "DocumentWindow",
    "Graphic",
    "HardwareSource",
    "Instrument",
    "Library"
}


def convert_to_facade(x, q):
    if isinstance(x, list):
        return [convert_to_facade(xx, q) for xx in x]
    if isinstance(x, tuple):
        return (convert_to_facade(xx, q) for xx in x)
    if isinstance(x, dict):
        return {k: convert_to_facade(v, q) for k, v in x.items()}
    class_name = x.__class__.__name__
    if class_name == "API_1":
        f = API(x, None)
        f._queue_task = q
        return f
    if class_name in all_classes:
        f = globals()[class_name](x, None)
        f._queue_task = q
        return f
    return x

def convert_from_facade(x):
    if isinstance(x, list):
        return [convert_from_facade(xx) for xx in x]
    if isinstance(x, tuple):
        return (convert_from_facade(xx) for xx in x)
    if isinstance(x, dict):
        return {k: convert_from_facade(v) for k, v in x.items()}
    class_name = x.__class__.__name__
    if class_name in all_classes:
        return getattr(x, "_proxy")
    return x

def queued(method):
    def queued(*args, **kwargs):
        result_ref = []
        exception_ref = []
        finished_event = threading.Event()
        target = args[0]
        def run():
            try:
                result_ref.append(convert_to_facade(method(*args, **kwargs), target._queue_task))
            except Exception as e:
                exception_ref.append(e)
            finally:
                finished_event.set()
        getattr(target, "_queue_task")(run)  # avoid type errors below
        finished_event.wait()
        if len(exception_ref) > 0:
            raise exception_ref[0]
        return result_ref[0]
    return queued

@queued
def call_method(target, method_name, *args, **kwargs):
    object = convert_from_facade(target)
    args = convert_from_facade(args)
    kwargs = convert_from_facade(kwargs)
    return getattr(object, method_name)(*args, **kwargs)

def call_threadsafe_method(target, method_name, *args, **kwargs):
    object = convert_from_facade(target)
    args = convert_from_facade(args)
    kwargs = convert_from_facade(kwargs)
    return getattr(object, method_name)(*args, **kwargs)

@queued
def get_property(target, property_name):
    return getattr(target._proxy, property_name)

@queued
def set_property(target, property_name, value):
    return setattr(target._proxy, property_name, value)

### the section below is copied from PlugIns/Connection/NionLib/nionlib/Classes.py


class Graphic:

    def __init__(self, proxy, specifier):
        self._proxy = proxy
        self.specifier = specifier

    def get_property(self, property):
        return call_method(self, 'get_property', property)

    def mask_xdata_with_shape(self, shape):
        return call_method(self, 'mask_xdata_with_shape', shape)

    def set_property(self, property, value):
        call_method(self, 'set_property', property, value)

    @property
    def angle(self):
        return get_property(self, 'angle')

    @angle.setter
    def angle(self, value):
        set_property(self, 'angle', value)

    @property
    def bounds(self):
        return get_property(self, 'bounds')

    @bounds.setter
    def bounds(self, value):
        set_property(self, 'bounds', value)

    @property
    def center(self):
        return get_property(self, 'center')

    @center.setter
    def center(self, value):
        set_property(self, 'center', value)

    @property
    def end(self):
        return get_property(self, 'end')

    @end.setter
    def end(self, value):
        set_property(self, 'end', value)

    @property
    def graphic_id(self):
        return get_property(self, 'graphic_id')

    @graphic_id.setter
    def graphic_id(self, value):
        set_property(self, 'graphic_id', value)

    @property
    def graphic_type(self):
        return get_property(self, 'graphic_type')

    @property
    def interval(self):
        return get_property(self, 'interval')

    @interval.setter
    def interval(self, value):
        set_property(self, 'interval', value)

    @property
    def label(self):
        return get_property(self, 'label')

    @label.setter
    def label(self, value):
        set_property(self, 'label', value)

    @property
    def position(self):
        return get_property(self, 'position')

    @position.setter
    def position(self, value):
        set_property(self, 'position', value)

    @property
    def region(self):
        return get_property(self, 'region')

    @property
    def size(self):
        return get_property(self, 'size')

    @size.setter
    def size(self, value):
        set_property(self, 'size', value)

    @property
    def start(self):
        return get_property(self, 'start')

    @start.setter
    def start(self, value):
        set_property(self, 'start', value)

    @property
    def type(self):
        return get_property(self, 'type')

    @property
    def uuid(self):
        return get_property(self, 'uuid')

    @property
    def vector(self):
        return get_property(self, 'vector')

    @vector.setter
    def vector(self, value):
        set_property(self, 'vector', value)


class DataItem:

    def __init__(self, proxy, specifier):
        self._proxy = proxy
        self.specifier = specifier

    def _repr_svg_(self):
        return call_method(self, 'data_item_to_svg')

    def add_channel_region(self, position):
        return call_method(self, 'add_channel_region', position)

    def add_ellipse_region(self, center_y, center_x, height, width):
        return call_method(self, 'add_ellipse_region', center_y, center_x, height, width)

    def add_interval_region(self, start, end):
        return call_method(self, 'add_interval_region', start, end)

    def add_line_region(self, start_y, start_x, end_y, end_x):
        return call_method(self, 'add_line_region', start_y, start_x, end_y, end_x)

    def add_point_region(self, y, x):
        return call_method(self, 'add_point_region', y, x)

    def add_rectangle_region(self, center_y, center_x, height, width):
        return call_method(self, 'add_rectangle_region', center_y, center_x, height, width)

    def delete_metadata_value(self, key):
        call_method(self, 'delete_metadata_value', key)

    def get_metadata_value(self, key):
        return call_method(self, 'get_metadata_value', key)

    def has_metadata_value(self, key):
        return call_method(self, 'has_metadata_value', key)

    def mask_xdata(self):
        return call_method(self, 'mask_xdata')

    def remove_region(self, graphic):
        call_method(self, 'remove_region', graphic)

    def set_data(self, data):
        call_method(self, 'set_data', data)

    def set_data_and_metadata(self, data_and_metadata):
        call_method(self, 'set_data_and_metadata', data_and_metadata)

    def set_dimensional_calibrations(self, dimensional_calibrations):
        call_method(self, 'set_dimensional_calibrations', dimensional_calibrations)

    def set_intensity_calibration(self, intensity_calibration):
        call_method(self, 'set_intensity_calibration', intensity_calibration)

    def set_metadata(self, metadata):
        call_method(self, 'set_metadata', metadata)

    def set_metadata_value(self, key, value):
        call_method(self, 'set_metadata_value', key, value)

    @property
    def created(self):
        return get_property(self, 'created')

    @property
    def data(self):
        return get_property(self, 'data')

    @data.setter
    def data(self, value):
        set_property(self, 'data', value)

    @property
    def data_and_metadata(self):
        return get_property(self, 'data_and_metadata')

    @property
    def dimensional_calibrations(self):
        return get_property(self, 'dimensional_calibrations')

    @property
    def display(self):
        return get_property(self, 'display')

    @property
    def display_xdata(self):
        return get_property(self, 'display_xdata')

    @property
    def graphics(self):
        return get_property(self, 'graphics')

    @property
    def intensity_calibration(self):
        return get_property(self, 'intensity_calibration')

    @property
    def metadata(self):
        return get_property(self, 'metadata')

    @property
    def modified(self):
        return get_property(self, 'modified')

    @property
    def regions(self):
        return get_property(self, 'regions')

    @property
    def title(self):
        return get_property(self, 'title')

    @title.setter
    def title(self, value):
        set_property(self, 'title', value)

    @property
    def uuid(self):
        return get_property(self, 'uuid')

    @property
    def xdata(self):
        return get_property(self, 'xdata')

    @xdata.setter
    def xdata(self, value):
        set_property(self, 'xdata', value)


class DisplayPanel:

    def __init__(self, proxy, specifier):
        self._proxy = proxy
        self.specifier = specifier

    def set_data_item(self, data_item):
        call_method(self, 'set_data_item', data_item)

    @property
    def data_item(self):
        return get_property(self, 'data_item')


class Display:

    def __init__(self, proxy, specifier):
        self._proxy = proxy
        self.specifier = specifier

    def get_graphic_by_id(self, graphic_id):
        return call_method(self, 'get_graphic_by_id', graphic_id)

    @property
    def data_item(self):
        return get_property(self, 'data_item')

    @property
    def display_type(self):
        return get_property(self, 'display_type')

    @display_type.setter
    def display_type(self, value):
        set_property(self, 'display_type', value)

    @property
    def graphics(self):
        return get_property(self, 'graphics')

    @property
    def selected_graphics(self):
        return get_property(self, 'selected_graphics')

    @property
    def uuid(self):
        return get_property(self, 'uuid')


class DataGroup:

    def __init__(self, proxy, specifier):
        self._proxy = proxy
        self.specifier = specifier

    def add_data_item(self, data_item):
        call_method(self, 'add_data_item', data_item)

    @property
    def uuid(self):
        return get_property(self, 'uuid')


class Library:

    def __init__(self, proxy, specifier):
        self._proxy = proxy
        self.specifier = specifier

    def copy_data_item(self, data_item):
        return call_method(self, 'copy_data_item', data_item)

    def create_data_item(self, title=None):
        return call_method(self, 'create_data_item', title=title)

    def create_data_item_from_data(self, data, title=None):
        return call_method(self, 'create_data_item_from_data', data, title=title)

    def create_data_item_from_data_and_metadata(self, data_and_metadata, title=None):
        return call_method(self, 'create_data_item_from_data_and_metadata', data_and_metadata, title=title)

    def data_ref_for_data_item(self, data_item):
        return call_method(self, 'data_ref_for_data_item', data_item)

    def delete_library_value(self, key):
        call_method(self, 'delete_library_value', key)

    def get_data_item_by_uuid(self, data_item_uuid):
        return call_method(self, 'get_data_item_by_uuid', data_item_uuid)

    def get_data_item_for_hardware_source(self, hardware_source, channel_id=None, processor_id=None, create_if_needed=False, large_format=False):
        return call_method(self, 'get_data_item_for_hardware_source', hardware_source, channel_id=channel_id, processor_id=processor_id, create_if_needed=create_if_needed, large_format=large_format)

    def get_dependent_data_items(self, data_item):
        return call_method(self, 'get_dependent_data_items', data_item)

    def get_graphic_by_uuid(self, graphic_uuid):
        return call_method(self, 'get_graphic_by_uuid', graphic_uuid)

    def get_library_value(self, key):
        return call_method(self, 'get_library_value', key)

    def get_or_create_data_group(self, title):
        return call_method(self, 'get_or_create_data_group', title)

    def get_source_data_items(self, data_item):
        return call_method(self, 'get_source_data_items', data_item)

    def has_library_value(self, key):
        return call_method(self, 'has_library_value', key)

    def set_library_value(self, key, value):
        call_method(self, 'set_library_value', key, value)

    def snapshot_data_item(self, data_item):
        return call_method(self, 'snapshot_data_item', data_item)

    @property
    def data_item_count(self):
        return get_property(self, 'data_item_count')

    @property
    def data_items(self):
        return get_property(self, 'data_items')

    @property
    def uuid(self):
        return get_property(self, 'uuid')


class DocumentWindow:

    def __init__(self, proxy, specifier):
        self._proxy = proxy
        self.specifier = specifier

    def add_data(self, data, title=None):
        return call_method(self, 'add_data', data, title=title)

    def create_data_item_from_data(self, data, title=None):
        return call_method(self, 'create_data_item_from_data', data, title=title)

    def create_data_item_from_data_and_metadata(self, data_and_metadata, title=None):
        return call_method(self, 'create_data_item_from_data_and_metadata', data_and_metadata, title=title)

    def display_data_item(self, data_item, source_display_panel=None, source_data_item=None):
        return call_method(self, 'display_data_item', data_item, source_display_panel=source_display_panel, source_data_item=source_data_item)

    def get_display_panel_by_id(self, identifier):
        return call_method(self, 'get_display_panel_by_id', identifier)

    def get_or_create_data_group(self, title):
        return call_method(self, 'get_or_create_data_group', title)

    def queue_task(self, fn):
        call_method(self, 'queue_task', fn)

    def show_confirmation_message_box(self, caption, accepted_fn, rejected_fn=None, accepted_text=None, rejected_text=None, display_rejected=False):
        call_method(self, 'show_confirmation_message_box', caption, accepted_fn, rejected_fn=rejected_fn, accepted_text=accepted_text, rejected_text=rejected_text, display_rejected=display_rejected)

    def show_get_string_message_box(self, caption, text, accepted_fn, rejected_fn=None, accepted_text=None, rejected_text=None):
        call_method(self, 'show_get_string_message_box', caption, text, accepted_fn, rejected_fn=rejected_fn, accepted_text=accepted_text, rejected_text=rejected_text)

    def show_modeless_dialog(self, item, handler=None):
        return call_method(self, 'show_modeless_dialog', item, handler=handler)

    @property
    def all_display_panels(self):
        return get_property(self, 'all_display_panels')

    @property
    def library(self):
        return get_property(self, 'library')

    @property
    def target_data_item(self):
        return get_property(self, 'target_data_item')

    @property
    def target_display(self):
        return get_property(self, 'target_display')


class Application:

    def __init__(self, proxy, specifier):
        self._proxy = proxy
        self.specifier = specifier

    @property
    def document_controllers(self):
        return get_property(self, 'document_controllers')

    @property
    def document_windows(self):
        return get_property(self, 'document_windows')

    @property
    def library(self):
        return get_property(self, 'library')


class API:

    def __init__(self, proxy, specifier):
        self._proxy = proxy
        self.specifier = specifier

    def create_calibration(self, offset=None, scale=None, units=None):
        return call_method(self, 'create_calibration', offset=offset, scale=scale, units=units)

    def create_data_and_metadata(self, data, intensity_calibration=None, dimensional_calibrations=None, metadata=None, timestamp=None, data_descriptor=None):
        return call_method(self, 'create_data_and_metadata', data, intensity_calibration=intensity_calibration, dimensional_calibrations=dimensional_calibrations, metadata=metadata, timestamp=timestamp, data_descriptor=data_descriptor)

    def create_data_and_metadata_from_data(self, data, intensity_calibration=None, dimensional_calibrations=None, metadata=None, timestamp=None):
        return call_method(self, 'create_data_and_metadata_from_data', data, intensity_calibration=intensity_calibration, dimensional_calibrations=dimensional_calibrations, metadata=metadata, timestamp=timestamp)

    def create_data_and_metadata_io_handler(self, io_handler_delegate):
        return call_method(self, 'create_data_and_metadata_io_handler', io_handler_delegate)

    def create_data_descriptor(self, is_sequence, collection_dimension_count, datum_dimension_count):
        return call_method(self, 'create_data_descriptor', is_sequence, collection_dimension_count, datum_dimension_count)

    def create_hardware_source(self, hardware_source_delegate):
        return call_method(self, 'create_hardware_source', hardware_source_delegate)

    def create_menu_item(self, menu_item_handler):
        return call_method(self, 'create_menu_item', menu_item_handler)

    def create_panel(self, panel_delegate):
        return call_method(self, 'create_panel', panel_delegate)

    def get_all_hardware_source_ids(self):
        return call_method(self, 'get_all_hardware_source_ids')

    def get_all_instrument_ids(self):
        return call_method(self, 'get_all_instrument_ids')

    def get_hardware_source_by_id(self, hardware_source_id, version):
        return call_method(self, 'get_hardware_source_by_id', hardware_source_id, version)

    def get_instrument_by_id(self, instrument_id, version):
        return call_method(self, 'get_instrument_by_id', instrument_id, version)

    def queue_task(self, fn):
        call_method(self, 'queue_task', fn)

    @property
    def application(self):
        return get_property(self, 'application')

    @property
    def library(self):
        return get_property(self, 'library')


class HardwareSource:

    def __init__(self, proxy, specifier):
        self._proxy = proxy
        self.specifier = specifier

    def abort_playing(self):
        call_method(self, 'abort_playing')

    def abort_recording(self):
        call_method(self, 'abort_recording')

    def close(self):
        call_method(self, 'close')

    def create_record_task(self, frame_parameters=None, channels_enabled=None):
        return call_method(self, 'create_record_task', frame_parameters=frame_parameters, channels_enabled=channels_enabled)

    def create_view_task(self, frame_parameters=None, channels_enabled=None, buffer_size=1):
        return call_method(self, 'create_view_task', frame_parameters=frame_parameters, channels_enabled=channels_enabled, buffer_size=buffer_size)

    def get_default_frame_parameters(self):
        return call_method(self, 'get_default_frame_parameters')

    def get_frame_parameters(self):
        return call_method(self, 'get_frame_parameters')

    def get_frame_parameters_for_profile_by_index(self, profile_index):
        return call_method(self, 'get_frame_parameters_for_profile_by_index', profile_index)

    def get_property_as_bool(self, name):
        return call_method(self, 'get_property_as_bool', name)

    def get_property_as_float(self, name):
        return call_method(self, 'get_property_as_float', name)

    def get_property_as_float_point(self, name):
        return call_method(self, 'get_property_as_float_point', name)

    def get_property_as_int(self, name):
        return call_method(self, 'get_property_as_int', name)

    def get_property_as_str(self, name):
        return call_method(self, 'get_property_as_str', name)

    def grab_next_to_finish(self, timeout=None):
        return call_threadsafe_method(self, 'grab_next_to_finish', timeout=timeout)

    def grab_next_to_start(self, frame_parameters=None, channels_enabled=None, timeout=None):
        return call_threadsafe_method(self, 'grab_next_to_start', frame_parameters=frame_parameters, channels_enabled=channels_enabled, timeout=timeout)

    def record(self, frame_parameters=None, channels_enabled=None, timeout=None):
        return call_threadsafe_method(self, 'record', frame_parameters=frame_parameters, channels_enabled=channels_enabled, timeout=timeout)

    def set_frame_parameters(self, frame_parameters):
        call_method(self, 'set_frame_parameters', frame_parameters)

    def set_frame_parameters_for_profile_by_index(self, profile_index, frame_parameters):
        call_method(self, 'set_frame_parameters_for_profile_by_index', profile_index, frame_parameters)

    def set_property_as_bool(self, name, value):
        call_threadsafe_method(self, 'set_property_as_bool', name, value)

    def set_property_as_float(self, name, value):
        call_threadsafe_method(self, 'set_property_as_float', name, value)

    def set_property_as_float_point(self, name, value):
        call_threadsafe_method(self, 'set_property_as_float_point', name, value)

    def set_property_as_int(self, name, value):
        call_threadsafe_method(self, 'set_property_as_int', name, value)

    def set_property_as_str(self, name, value):
        call_threadsafe_method(self, 'set_property_as_str', name, value)

    def start_playing(self, frame_parameters=None, channels_enabled=None):
        call_method(self, 'start_playing', frame_parameters=frame_parameters, channels_enabled=channels_enabled)

    def start_recording(self, frame_parameters=None, channels_enabled=None):
        return call_method(self, 'start_recording', frame_parameters=frame_parameters, channels_enabled=channels_enabled)

    def stop_playing(self):
        call_method(self, 'stop_playing')

    @property
    def is_playing(self):
        return get_property(self, 'is_playing')

    @property
    def is_recording(self):
        return get_property(self, 'is_recording')

    @property
    def profile_index(self):
        return get_property(self, 'profile_index')

    @profile_index.setter
    def profile_index(self, value):
        set_property(self, 'profile_index', value)


class Instrument:

    def __init__(self, proxy, specifier):
        self._proxy = proxy
        self.specifier = specifier

    def close(self):
        call_method(self, 'close')

    def get_control_output(self, name):
        return call_method(self, 'get_control_output', name)

    def get_control_state(self, name):
        return call_method(self, 'get_control_state', name)

    def get_property_as_bool(self, name):
        return call_method(self, 'get_property_as_bool', name)

    def get_property_as_float(self, name):
        return call_method(self, 'get_property_as_float', name)

    def get_property_as_float_point(self, name):
        return call_method(self, 'get_property_as_float_point', name)

    def get_property_as_int(self, name):
        return call_method(self, 'get_property_as_int', name)

    def get_property_as_str(self, name):
        return call_method(self, 'get_property_as_str', name)

    def set_control_output(self, name, value, *, options=None):
        call_method(self, 'set_control_output', name, value, options=options)

    def set_property_as_bool(self, name, value):
        call_method(self, 'set_property_as_bool', name, value)

    def set_property_as_float(self, name, value):
        call_method(self, 'set_property_as_float', name, value)

    def set_property_as_float_point(self, name, value):
        call_method(self, 'set_property_as_float_point', name, value)

    def set_property_as_int(self, name, value):
        call_method(self, 'set_property_as_int', name, value)

    def set_property_as_str(self, name, value):
        call_method(self, 'set_property_as_str', name, value)



# standard libraries
import collections
import copy
import importlib
import importlib.util
import inspect
import json
import logging
import os
import pkgutil
import re
import sys
import traceback
import unittest

from nion.swift.model import Utility
from nion.ui import Declarative


__modules = []
__test_suites = []


class RequirementsException(Exception):
    """An exception for when a plug-in can't load because it can't meet the necessary requirements."""
    def __init__(self, reason):
        self.reason = reason


api_broker_fn = None

def register_api_broker_fn(new_api_broker_fn):
    global api_broker_fn
    api_broker_fn = new_api_broker_fn

class APIBroker:
    def get_api(self, *args, **kwargs):
        global api_broker_fn
        return api_broker_fn(*args, **kwargs)
    def get_ui(self, version):
        actual_version = "1.0.0"
        if Utility.compare_versions(version, actual_version) > 0:
            raise NotImplementedError("API requested version %s is greater than %s." % (version, actual_version))
        return Declarative.DeclarativeUI()

extensions = []

def load_plug_in(module_path: str, module_name: str):
    try:
        # First load the module.
        module = importlib.import_module(module_name)
        # Now scan through the module and look for extensions and tests.
        tests = []
        for member in inspect.getmembers(module):
            if inspect.ismodule(member[1]):
                if member[0].endswith("_test"):
                    for maybe_a_class in inspect.getmembers(member[1]):
                        if inspect.isclass(maybe_a_class[1]):
                            if maybe_a_class[0].startswith("Test"):
                                # It is a class... add it to the test suite.
                                tests.append(maybe_a_class[0])
                                cls = getattr(member[1], maybe_a_class[0])
                                __test_suites.append(unittest.TestLoader().loadTestsFromTestCase(cls))
                for maybe_a_class in inspect.getmembers(member[1]):
                    if inspect.isclass(maybe_a_class[1]):
                        if maybe_a_class[0].endswith("Extension"):
                            cls = getattr(member[1], maybe_a_class[0])
                            extension_id = getattr(cls, "extension_id", None)
                            if extension_id:
                                extensions.append(cls(APIBroker()))
            if inspect.isclass(member[1]) and member[0].endswith("Extension"):
                cls = getattr(module, member[0])
                extension_id = getattr(cls, "extension_id", None)
                if extension_id:
                    extensions.append(cls(APIBroker()))
        plugin_loaded_str = "Plug-in '" + module_name + "' loaded (" + module_path + ")."
        list_of_tests_str = " Tests: " + ",".join(tests) if len(tests) > 0 else ""
        logging.info(plugin_loaded_str + list_of_tests_str)
        return module
    except RequirementsException as e:
        logging.info("Plug-in '" + module_name + "' NOT loaded. %s (" + module_path + ")", e.reason)
    except Exception as e:
        logging.info("Plug-in '" + module_name + "' NOT loaded (" + module_path + ").")
        logging.info(traceback.format_exc())
        logging.info("--------")
    return None


class ModuleAdapter:
    def __init__(self, package_name, module_info):
        self.module_name = package_name + "." + module_info.name
        self.module_path = package_name
        self.manifest_path = "N/A"
        self.manifest = dict()
        path = getattr(module_info.module_finder, 'path', None)
        if path:
            self.manifest_path = os.path.join(path, module_info.name, "manifest.json")
            if os.path.exists(self.manifest_path):
                try:
                    with open(self.manifest_path) as f:
                        self.manifest.update(json.load(f))
                except Exception as e:
                    logging.info("Cannot read manifest file from %s", self.manifest_path)
                    logging.info(e)
        get_data = getattr(module_info.module_finder, 'get_data', None)
        if get_data:
            try:
                json_data = get_data(os.path.join(package_name, module_info.name, "manifest.json"))
                self.manifest.update(json.loads(json_data))
            except Exception as e:
                logging.info("Cannot read manifest file from %s", module_info.module_finder.archive)
                logging.info(e)
        self.manifest.setdefault("name", self.module_name)
        self.manifest.setdefault("identifier", self.module_name)
        self.manifest.setdefault("version", "0.0.0")

    def load(self):
        return load_plug_in(self.module_path, self.module_name)


class PlugInAdapter:
    def __init__(self, directory, relative_path):
        plugin_dir = os.path.join(directory, relative_path)
        self.manifest_path = os.path.join(plugin_dir, "manifest.json")
        self.module_name = relative_path
        self.module_path = directory
        self.manifest = None
        if os.path.exists(self.manifest_path):
            try:
                with open(self.manifest_path) as f:
                    self.manifest = json.load(f)
            except Exception as e:
                logging.info("Cannot read manifest file from %s", self.manifest_path)
                logging.info(e)

    def load(self):
        return load_plug_in(self.module_path, self.module_name)


def load_plug_ins(app, root_dir):
    """Load plug-ins."""
    global extensions

    ui = app.ui

    # a list of directories in which sub-directories PlugIns will be searched.
    subdirectories = []

    # the default location is where the directory main packages are located.
    if root_dir:
        subdirectories.append(root_dir)

    # also search the default data location; create directory there if it doesn't exist to make it easier for user.
    # default data location will be application specific.
    data_location = ui.get_data_location()
    if data_location is not None:
        subdirectories.append(data_location)
        # create directories here if they don't exist
        plugins_dir = os.path.abspath(os.path.join(data_location, "PlugIns"))
        if not os.path.exists(plugins_dir):
            logging.info("Creating plug-ins directory %s", plugins_dir)
            os.makedirs(plugins_dir)

    # search the Nion/Swift subdirectory of the default document location too,
    # but don't create directories here - avoid polluting user visible directories.
    document_location = ui.get_document_location()
    if document_location is not None:
        subdirectories.append(os.path.join(document_location, "Nion", "Swift"))
        # do not create them in documents if they don't exist. this location is optional.

    # build a list of directories that will be loaded as plug-ins.
    PlugInDir = collections.namedtuple("PlugInDir", ["directory", "relative_path"])
    plugin_dirs = list()

    # track directories that have already been searched.
    seen_plugin_dirs = list()

    # for each subdirectory, look in PlugIns for sub-directories that represent the plug-ins.
    for subdirectory in subdirectories:
        plugins_dir = os.path.abspath(os.path.join(subdirectory, "PlugIns"))

        if os.path.exists(plugins_dir) and not plugins_dir in seen_plugin_dirs:
            logging.info("Loading plug-ins from %s", plugins_dir)

            # add the PlugIns directory to the system import path.
            sys.path.append(plugins_dir)

            # now build a list of sub-directories representing plug-ins within plugins_dir.
            sorted_relative_paths = sorted([d for d in os.listdir(plugins_dir) if os.path.isdir(os.path.join(plugins_dir, d))])
            plugin_dirs.extend([PlugInDir(plugins_dir, sorted_relative_path) for sorted_relative_path in sorted_relative_paths])

            # mark plugins_dir as 'seen' to avoid search it twice.
            seen_plugin_dirs.append(plugins_dir)
        else:
            logging.info("NOT Loading plug-ins from %s (missing)", plugins_dir)

    version_map = dict()
    module_exists_map = dict()

    plugin_adapters = list()

    import nionswift_plugin
    for module_info in pkgutil.iter_modules(nionswift_plugin.__path__):
        plugin_adapters.append(ModuleAdapter(nionswift_plugin.__name__, module_info))

    for directory, relative_path in plugin_dirs:
        plugin_adapters.append(PlugInAdapter(directory, relative_path))

    progress = True
    while progress:
        progress = False
        plugin_adapters_copy = copy.deepcopy(plugin_adapters)
        plugin_adapters = list()
        for plugin_adapter in plugin_adapters_copy:
            manifest_path = plugin_adapter.manifest_path
            manifest = plugin_adapter.manifest
            if manifest:
                manifest_valid = True
                if not "name" in manifest:
                    logging.info("Invalid manifest (missing 'name'): %s", manifest_path)
                    manifest_valid = False
                if not "identifier" in manifest:
                    logging.info("Invalid manifest (missing 'identifier'): %s", manifest_path)
                    manifest_valid = False
                if "identifier" in manifest and not re.match("[_\-a-zA-Z][_\-a-zA-Z0-9.]*$", manifest["identifier"]):
                    logging.info("Invalid manifest (invalid 'identifier': '%s'): %s", manifest["identifier"], manifest_path)
                    manifest_valid = False
                if not "version" in manifest:
                    logging.info("Invalid manifest (missing 'version'): %s", manifest_path)
                    manifest_valid = False
                if "requires" in manifest and not isinstance(manifest["requires"], list):
                    logging.info("Invalid manifest ('requires' not a list): %s", manifest_path)
                    manifest_valid = False
                if not manifest_valid:
                    continue
                for module in manifest.get("modules", list()):
                    if module in module_exists_map:
                        module_exists = module_exists_map.get(module)
                    else:
                        module_exists = importlib.util.find_spec(module) is not None
                        module_exists_map[module] = module_exists
                    if not module_exists:
                        logging.info("Plug-in '" + plugin_adapter.module_name + "' NOT loaded (" + plugin_adapter.module_path + ").")
                        logging.info("Cannot satisfy requirement (%s): %s", module, manifest_path)
                        manifest_valid = False
                        break
                for requirement in manifest.get("requires", list()):
                    # TODO: see https://packaging.pypa.io/en/latest/
                    requirement_components = requirement.split()
                    if len(requirement_components) != 3 or requirement_components[1] != "~=":
                        logging.info("Invalid manifest (requirement '%s' invalid): %s", requirement, manifest_path)
                        manifest_valid = False
                        break
                    identifier, operator, version_specifier = requirement_components[0], requirement_components[1], requirement_components[2]
                    if identifier in version_map:
                        if Utility.compare_versions("~" + version_specifier, version_map[identifier]) != 0:
                            logging.info("Plug-in '" + plugin_adapter.module_name + "' NOT loaded (" + plugin_adapter.module_path + ").")
                            logging.info("Cannot satisfy requirement (%s): %s", requirement, manifest_path)
                            manifest_valid = False
                            break
                    else:
                        # requirements not loaded yet; add back to plugin_adapters, but don't mark progress since nothing was loaded.
                        logging.info("Plug-in '" + plugin_adapter.module_name + "' delayed (%s) (" + plugin_adapter.module_path + ").", requirement)
                        plugin_adapters.append(plugin_adapter)
                        manifest_valid = False
                        break
                if not manifest_valid:
                    continue
                version_map[manifest["identifier"]] = manifest["version"]
            # read the manifests, if any
            # repeat loop of plug-ins until no plug-ins left in the list
            #   if all dependencies satisfied for a plug-in, load it
            #   otherwise defer until next round
            #   stop if no plug-ins loaded in the round
            #   count on the user to have correct dependencies
            module = plugin_adapter.load()
            if module:
                __modules.append(module)
            progress = True
    for plugin_adapter in plugin_adapters:
        logging.info("Plug-in '" + plugin_adapter.module_name + "' NOT loaded (requirements) (" + plugin_adapter.module_path + ").")

    notify_modules("run")

def unload_plug_ins():
    global extensions
    for extension in reversed(extensions):
        try:
            extension.close()
        except Exception as e:
            logging.info(traceback.format_exc())

    extensions = []

def notify_modules(method_name, *args, **kwargs):
    for module in __modules:
        for member in inspect.getmembers(module):
            if inspect.isfunction(member[1]) and member[0] == method_name:
                try:
                    member[1](*args, **kwargs)
                except Exception as e:
                    logging.info("Plug-in '" + str(module) + "' exception during '" + method_name + "'.")
                    logging.info(traceback.format_exc())
                    logging.info("--------")


def append_test_suites(suites):
    __test_suites.extend(suites)


def test_suites():
    return __test_suites



# standard libraries
import contextlib
import copy
import unittest

# third party libraries
import numpy

# local libraries
from nion.swift import Application
from nion.swift import Facade
from nion.swift.model import DataItem
from nion.swift.model import DisplayItem
from nion.swift.model import DocumentModel
from nion.ui import TestUI


Facade.initialize()


class TestDisplayItemClass(unittest.TestCase):

    def setUp(self):
        self.app = Application.Application(TestUI.UserInterface(), set_global=False)

    def tearDown(self):
        pass

    def test_display_item_with_multiple_display_data_channels_has_sensible_properties(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            data_item2 = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            document_model.append_data_item(data_item2, False)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.append_display_data_channel(DisplayItem.DisplayDataChannel(data_item=data_item2))
            self.assertIsNotNone(display_item.size_and_data_format_as_string)
            self.assertIsNotNone(display_item.date_for_sorting)
            self.assertIsNotNone(display_item.date_for_sorting_local_as_string)
            self.assertIsNotNone(display_item.status_str)
            self.assertIsNotNone(display_item.used_display_type)

    def test_display_item_snapshot_and_copy_preserve_display_type(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.display_type = "line_plot"
            snapshot_display_item = display_item.snapshot()
            self.assertEqual("line_plot", snapshot_display_item.display_type)
            copy_display_item = copy.deepcopy(display_item)
            self.assertEqual("line_plot", copy_display_item.display_type)

    def test_appending_display_data_channel_does_nothing_if_display_data_channel_already_exists(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item1 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item2)
            display_item = document_model.get_display_item_for_data_item(data_item1)
            self.assertEqual(1, len(display_item.display_data_channels))
            display_item.append_display_data_channel_for_data_item(data_item2)
            self.assertEqual(2, len(display_item.display_data_channels))
            display_item.append_display_data_channel_for_data_item(data_item2)
            self.assertEqual(2, len(display_item.display_data_channels))

    def test_appending_display_data_channel_adds_layer(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item1 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item2)
            display_item = document_model.get_display_item_for_data_item(data_item1)
            self.assertEqual(1, len(display_item.display_layers))
            display_item.append_display_data_channel_for_data_item(data_item2)
            self.assertEqual(2, len(display_item.display_data_channels))

    def test_removing_data_item_updates_display_layer_data_indexes(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item1 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item2)
            display_item = document_model.get_display_item_for_data_item(data_item1)
            display_item.append_display_data_channel_for_data_item(data_item2)
            display_item._set_display_layer_property(0, "ref", "A")
            display_item._set_display_layer_property(1, "ref", "B")
            self.assertEqual(2, len(display_item.display_data_channels))
            self.assertEqual(0, display_item.get_display_layer_property(0, "data_index"))
            self.assertEqual(1, display_item.get_display_layer_property(1, "data_index"))
            document_model.remove_data_item(data_item1)
            self.assertEqual(1, len(display_item.display_layers))
            self.assertEqual(0, display_item.get_display_layer_property(0, "data_index"))
            self.assertEqual("B", display_item.get_display_layer_property(0, "ref"))

    def test_inserting_display_data_channel_updates_display_layer_data_indexes(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item1 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item1)
            data_item2 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item2)
            data_item3 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item3)
            display_item = document_model.get_display_item_for_data_item(data_item1)
            display_item.append_display_data_channel_for_data_item(data_item2)
            display_item._set_display_layer_property(0, "ref", "A")
            display_item._set_display_layer_property(1, "ref", "B")
            self.assertEqual(2, len(display_item.display_data_channels))
            self.assertEqual(0, display_item.get_display_layer_property(0, "data_index"))
            self.assertEqual(1, display_item.get_display_layer_property(1, "data_index"))
            display_item.insert_display_data_channel(1, DisplayItem.DisplayDataChannel(data_item=data_item3))
            self.assertEqual(3, len(display_item.display_data_channels))
            self.assertEqual(0, display_item.get_display_layer_property(0, "data_index"))
            self.assertEqual(2, display_item.get_display_layer_property(1, "data_index"))
            self.assertEqual("A", display_item.get_display_layer_property(0, "ref"))
            self.assertEqual("B", display_item.get_display_layer_property(1, "ref"))

    def test_copy_display_item_should_copy_all_display_data_channels_and_layers(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            data_item2 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item)
            document_model.append_data_item(data_item2, False)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.append_display_data_channel_for_data_item(data_item2)
            self.assertEqual(1, len(document_model.display_items))
            self.assertEqual(2, len(document_model.data_items))
            self.assertEqual(2, len(display_item.data_items))
            self.assertEqual(2, len(display_item.display_layers))
            display_item_copy = document_model.get_display_item_copy_new(display_item)
            self.assertEqual(2, len(document_model.data_items))
            self.assertEqual(2, len(document_model.display_items))
            self.assertEqual(2, len(display_item_copy.data_items))
            self.assertEqual(2, len(display_item_copy.display_layers))
            self.assertEqual(display_item.display_layers, display_item_copy.display_layers)

    def test_snapshot_display_item_with_data_item_and_multiple_display_layers(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((3, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.display_type = "line_plot"
            display_item.display_layers = [
                {"data_index": 0, "data_row": 2, "label": "Signal"},
                {"data_index": 0, "data_row": 1, "label": "Background"},
                {"data_index": 0, "data_row": 0, "label": "Data"},
            ]
            self.assertEqual(1, len(document_model.display_items))
            self.assertEqual(1, len(document_model.data_items))
            self.assertEqual(1, len(display_item.data_items))
            self.assertEqual(3, len(display_item.display_layers))
            display_item_copy = document_model.get_display_item_snapshot_new(display_item)
            self.assertEqual(2, len(document_model.display_items))
            self.assertEqual(2, len(document_model.data_items))
            self.assertEqual(1, len(display_item_copy.data_items))
            self.assertEqual(3, len(display_item_copy.display_layers))
            self.assertEqual(display_item.display_layers, display_item_copy.display_layers)

    def test_add_layer_to_line_plot_with_auto_layer_color_sets_both_colors(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            data_item2 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item)
            document_model.append_data_item(data_item2)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.remove_display_layer(0)
            self.assertEqual(0, len(display_item.display_layers))
            display_item.append_display_data_channel_for_data_item(data_item2)
            self.assertEqual(2, len(display_item.display_layers))
            self.assertIn("data_index", display_item.display_layers[0])
            self.assertIn("data_index", display_item.display_layers[1])
            self.assertIn("fill_color", display_item.display_layers[0])
            self.assertIn("fill_color", display_item.display_layers[1])

    def test_second_layer_to_line_plot_enables_caption(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            data_item2 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            data_item3 = DataItem.DataItem(numpy.zeros((8,), numpy.uint32))
            document_model.append_data_item(data_item)
            document_model.append_data_item(data_item2)
            document_model.append_data_item(data_item3)
            display_item = document_model.get_display_item_for_data_item(data_item)
            self.assertIsNone(display_item.get_display_property("legend_position"))
            # check that legend is automatically enabled for 2nd layer
            display_item.append_display_data_channel_for_data_item(data_item2)
            self.assertEqual("top-right", display_item.get_display_property("legend_position"))
            # check that legend is not automatically enabled for 3rd layer
            display_item.set_display_property("legend_position", None)
            display_item.append_display_data_channel_for_data_item(data_item3)
            self.assertIsNone(display_item.get_display_property("legend_position"))

    # test_transaction_does_not_cascade_to_data_item_refs
    # test_increment_data_ref_counts_cascades_to_data_item_refs
    # test_adding_data_item_twice_to_composite_item_fails
    # test_composition_item_starts_drag_with_composition_item_mime_data
    # test_composite_library_item_produces_composite_display
    # test_changing_display_type_of_composite_updates_displays_in_canvas_item
    # test_changing_display_type_of_child_updates_composite_display
    # test_composite_item_deletes_cleanly_when_displayed
    # test_delete_composite_cascade_delete_works
    # test_creating_r_var_on_composite_item
    # test_creating_r_var_on_library_items
    # test_transaction_on_composite_display_propagates_to_dependents
    # test_composite_item_deletes_children_when_deleted
    # test_undelete_composite_item
    # test_composite_line_plot_initializes_properly
    # test_composite_line_plot_calculates_calibrated_data_of_two_data_items_with_same_units_but_different_scales_properly
    # test_composite_line_plot_handles_drawing_with_fixed_y_scale_and_without_data
    # test_composite_line_plot_handles_first_components_without_data
    # test_multi_line_plot_without_calibration_does_not_display_any_line_graphs
    # test_multi_line_plot_handles_calibrated_vs_uncalibrated_display
    # test_delete_and_undelete_from_memory_storage_system_restores_composite_item_after_reload
    # test_data_item_with_references_to_another_data_item_reloads
    # test_composite_library_item_reloads_metadata
    # test_composite_data_item_saves_to_file_storage
    # test_composition_display_thumbnail_source_produces_library_item_mime_data


if __name__ == '__main__':
    unittest.main()



"""
See color maps:
    https://sciviscolor.org/
    http://www.kennethmoreland.com/color-advice/
    https://datascience.lanl.gov/colormaps.html
"""

import collections
import colorsys
import gettext
import json
import numpy
import math
import os
import pkgutil
import re
import typing
import xml.etree.ElementTree as ET

_ = gettext.gettext


def interpolate_colors(array: numpy.ndarray, x: int) -> numpy.ndarray:
    """
    Creates a color map for values in array
    :param array: color map to interpolate
    :param x: number of colors
    :return: interpolated color map
    """
    out_array = []
    for i in range(x):
        if i % (x / (len(array) - 1)) == 0:
            index = i / (x / (len(array) - 1))
            out_array.append(array[int(index)])
        else:
            start_marker = array[math.floor(i / (x / (len(array) - 1)))]
            stop_marker = array[math.ceil(i / (x / (len(array) - 1)))]
            interp_amount = i % (x / (len(array) - 1)) / (x / (len(array) - 1))
            interp_color = numpy.rint(start_marker + ((stop_marker - start_marker) * interp_amount))
            out_array.append(interp_color)
    out_array[-1] = array[-1]
    return numpy.array(out_array).astype(numpy.uint8)


def generate_lookup_array_from_points(points: typing.List[typing.Dict], n: int) -> numpy.ndarray:
    assert points[0]["x"] == 0.0
    assert points[-1]["x"] == 1.0
    out_array = []
    last_ix = None
    last_rgb = None
    for point in points:
        if "rgb" in point:
            r, g, b = point["rgb"]
        else:
            r = round(point["r"] * 255)
            g = round(point["g"] * 255)
            b = round(point["b"] * 255)
        x = point["x"]
        assert 0 <= r <= 255
        assert 0 <= g <= 255
        assert 0 <= b <= 255
        assert 0 <= x <= 1
        rgb = numpy.array([b, g, r])
        ix = int(math.floor(x * (n - 1)))
        if last_ix is None:
            out_array.append(numpy.copy(rgb))
        elif ix > last_ix:
            amount = (rgb - last_rgb) / (ix - last_ix)
            for x in range(ix - last_ix):
                out_array.append(numpy.rint(last_rgb + amount * (x + 1)))
        else:
            assert ix >= last_ix
        last_ix = ix
        last_rgb = numpy.copy(rgb)
    return numpy.array(out_array).astype(numpy.uint8)


def generate_lookup_array_grayscale():
    out_list = []
    for i in range(256):
        out_list.append([i, i, i])
    return numpy.array(out_list)

lookup_arrays = {
    'magma':     [[0, 0, 0],
                  [127, 0, 127],
                  [96, 96, 255],
                  [64, 192, 255],
                  [223, 255, 255]],
    'viridis':   [[32, 32, 64],
                  [160, 160, 64],
                  [208, 208, 0],
                  [144, 255, 32],
                  [0, 255, 255]],
    'plasma':    [[127, 0, 0],
                  [127, 0, 127],
                  [127, 127, 255],
                  [127, 255, 255],
                  [195, 255, 255]],
    'ice':       [[0, 0, 0],
                  [127, 0, 0],
                  [255, 127, 0],
                  [255, 255, 127],
                  [255, 255, 255]]
}

def generate_lookup_array(color_map_id: str) -> numpy.array:
    return interpolate_colors(numpy.array(lookup_arrays[color_map_id]), 256)

def generate_lookup_array_hsv() -> numpy.array:
    result_array = []
    for lookup_value in range(256):
        color_values = [x * 255 for x in colorsys.hsv_to_rgb(lookup_value / 300, 1.0, 1.0)]
        result_array.append(color_values)
    return numpy.array(result_array).astype(int)

ColorMap = collections.namedtuple("ColorMap", ["name", "data"])

color_maps = dict()

color_maps["grayscale"] = ColorMap(_("Grayscale"), generate_lookup_array_grayscale())
color_maps["magma"] = ColorMap(_("Magma"), generate_lookup_array('magma'))
color_maps["hsv"] = ColorMap(_("HSV"), generate_lookup_array_hsv())
color_maps["viridis"] = ColorMap(_("Viridis"), generate_lookup_array('viridis'))
color_maps["plasma"] = ColorMap(_("Plasma"), generate_lookup_array('plasma'))
color_maps["ice"] = ColorMap(_("Ice"), generate_lookup_array('ice'))

def load_color_maps(color_maps_dir) -> None:
    for root, dirs, files in os.walk(color_maps_dir):
        for file in files:
            if not file.startswith("."):
                try:
                    if file.endswith(".json"):
                        with open(os.path.join(root, file), "r") as f:
                            color_map_json = json.load(f)
                            color_maps[color_map_json["id"]] = ColorMap(color_map_json["name"], generate_lookup_array_from_points(color_map_json["points"], 256))
                    elif file.endswith(".xml"):
                        tree = ET.parse(os.path.join(root, file))
                        assert tree.getroot().tag == "ColorMaps"
                        color_map_tree = list(tree.getroot())[0]
                        assert color_map_tree.tag == "ColorMap"
                        raw_points = [point_tree.attrib for point_tree in color_map_tree]
                        points = list()
                        for raw_point in raw_points:
                            if "x" in raw_point:
                                points.append({"x": float(raw_point['x']), "r": float(raw_point['r']), "g": float(raw_point['g']), "b": float(raw_point['b'])})
                        name = file[:-4]
                        color_map_id = name.lower()
                        color_map_id = re.sub(r"[^\w\s]", '', color_map_id)
                        color_map_id = re.sub(r"\s+", '-', color_map_id)
                        color_maps[color_map_id] = ColorMap(name, generate_lookup_array_from_points(points, 256))
                        """
                        # this section can be used to generate .json from .xml color tables
                        points2 = [{"x": point['x'], "rgb": [round(point['r'] * 255), round(point['g'] * 255), round(point['b'] * 255)]} for point in points]
                        s = ""
                        s += '{' + '\n'
                        s += f'  "id": "{color_map_id}",' + '\n'
                        s += f'  "name": "{name}",' + '\n'
                        s += '  "points": [' + '\n'
                        bro = '{'
                        brc = '}'
                        for point2 in points2[:-1]:
                            s += f'    {bro}"x": {point2["x"]}, "rgb": [{point2["rgb"][0]}, {point2["rgb"][1]}, {point2["rgb"][2]}]{brc},' + '\n'
                        point2 = points2[-1]
                        s += f'    {bro}"x": {point2["x"]}, "rgb": [{point2["rgb"][0]}, {point2["rgb"][1]}, {point2["rgb"][2]}]{brc}' + '\n'
                        s += '  ]' + '\n'
                        s += '}' + '\n'
                        print(s)
                        # d = {"id": color_map_id, "name": name, "points": points2}
                        # print(json.dumps(d, indent=2))
                        """
                except Exception as e:
                    import traceback
                    traceback.print_exc()

def load_color_map_resource(resource_path: str) -> None:
    color_map_json = json.loads(pkgutil.get_data(__name__, resource_path))
    color_maps[color_map_json["id"]] = ColorMap(color_map_json["name"], generate_lookup_array_from_points(color_map_json["points"], 256))

load_color_map_resource("resources/color_maps/black_body.json")
load_color_map_resource("resources/color_maps/extended_black_body.json")
load_color_map_resource("resources/color_maps/extended_kindlmann.json")
load_color_map_resource("resources/color_maps/kindlmann.json")

def get_color_map_data_by_id(color_map_id: str) -> numpy.ndarray:
    return color_maps.get(color_map_id, color_maps["grayscale"]).data
